# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum


class EventHandlerKey(Enum):
    """Key bound to a `BaseEventHandler` plugin class.

    :ivar FIXED_CENTER_FREQUENCY: Postprocess data capture at a fixed center frequency.
    :ivar SWEPT_CENTER_FREQUENCY: Postprocess data capture where the center frequency is continually sweeping
    in fixed increments.
    """

    FIXED_CENTER_FREQUENCY = "fixed_center_frequency"
    SWEPT_CENTER_FREQUENCY = "swept_center_frequency"



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from spectre_core.capture_configs import Parameters, PName
from spectre_core.config import get_batches_dir_path
from ._gr import spectre_top_block

from gnuradio import spectre
from gnuradio import sdrplay3


class tuner_1_fixed_center_frequency(spectre_top_block):
    def flowgraph(self, tag: str, parameters: Parameters) -> None:
        # Unpack the capture config parameters
        sample_rate = parameters.get_parameter_value(PName.SAMPLE_RATE)
        batch_size = parameters.get_parameter_value(PName.BATCH_SIZE)
        center_freq = parameters.get_parameter_value(PName.CENTER_FREQUENCY)
        bandwidth = parameters.get_parameter_value(PName.BANDWIDTH)
        if_gain = parameters.get_parameter_value(PName.IF_GAIN)
        rf_gain = parameters.get_parameter_value(PName.RF_GAIN)

        # Blocks
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            get_batches_dir_path(), tag, batch_size, sample_rate
        )
        self.sdrplay3_rspduo = sdrplay3.rspduo(
            "",
            rspduo_mode="Single Tuner",
            antenna="Tuner 1 50 ohm",
            stream_args=sdrplay3.stream_args(output_type="fc32", channels_size=1),
        )
        self.sdrplay3_rspduo.set_sample_rate(sample_rate)
        self.sdrplay3_rspduo.set_center_freq(center_freq)
        self.sdrplay3_rspduo.set_bandwidth(bandwidth)
        self.sdrplay3_rspduo.set_antenna("Tuner 1 50 ohm")
        self.sdrplay3_rspduo.set_gain_mode(False)
        self.sdrplay3_rspduo.set_gain(if_gain, "IF")
        self.sdrplay3_rspduo.set_gain(rf_gain, "RF")
        self.sdrplay3_rspduo.set_freq_corr(0)
        self.sdrplay3_rspduo.set_dc_offset_mode(False)
        self.sdrplay3_rspduo.set_iq_balance_mode(False)
        self.sdrplay3_rspduo.set_agc_setpoint(-30)
        self.sdrplay3_rspduo.set_rf_notch_filter(False)
        self.sdrplay3_rspduo.set_dab_notch_filter(False)
        self.sdrplay3_rspduo.set_am_notch_filter(False)
        self.sdrplay3_rspduo.set_biasT(False)
        self.sdrplay3_rspduo.set_debug_mode(False)
        self.sdrplay3_rspduo.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo.set_show_gain_changes(False)

        # Connections
        self.connect((self.sdrplay3_rspduo, 0), (self.spectre_batched_file_sink, 0))


class tuner_2_fixed_center_frequency(spectre_top_block):
    def flowgraph(self, tag: str, parameters: Parameters) -> None:

        # Unpack the capture config parameters
        sample_rate = parameters.get_parameter_value(PName.SAMPLE_RATE)
        batch_size = parameters.get_parameter_value(PName.BATCH_SIZE)
        center_freq = parameters.get_parameter_value(PName.CENTER_FREQUENCY)
        bandwidth = parameters.get_parameter_value(PName.BANDWIDTH)
        if_gain = parameters.get_parameter_value(PName.IF_GAIN)
        rf_gain = parameters.get_parameter_value(PName.RF_GAIN)

        # Blocks
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            get_batches_dir_path(), tag, batch_size, sample_rate
        )
        self.sdrplay3_rspduo = sdrplay3.rspduo(
            "",
            rspduo_mode="Single Tuner",
            antenna="Tuner 2 50 ohm",
            stream_args=sdrplay3.stream_args(output_type="fc32", channels_size=1),
        )
        self.sdrplay3_rspduo.set_sample_rate(sample_rate)
        self.sdrplay3_rspduo.set_center_freq(center_freq)
        self.sdrplay3_rspduo.set_bandwidth(bandwidth)
        self.sdrplay3_rspduo.set_antenna("Tuner 2 50 ohm")
        self.sdrplay3_rspduo.set_gain_mode(False)
        self.sdrplay3_rspduo.set_gain(if_gain, "IF")
        self.sdrplay3_rspduo.set_gain(rf_gain, "RF", False)
        self.sdrplay3_rspduo.set_freq_corr(0)
        self.sdrplay3_rspduo.set_dc_offset_mode(False)
        self.sdrplay3_rspduo.set_iq_balance_mode(False)
        self.sdrplay3_rspduo.set_agc_setpoint(-30)
        self.sdrplay3_rspduo.set_rf_notch_filter(False)
        self.sdrplay3_rspduo.set_dab_notch_filter(False)
        self.sdrplay3_rspduo.set_am_notch_filter(False)
        self.sdrplay3_rspduo.set_biasT(False)
        self.sdrplay3_rspduo.set_stream_tags(False)
        self.sdrplay3_rspduo.set_debug_mode(False)
        self.sdrplay3_rspduo.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo.set_show_gain_changes(False)

        # Connections
        self.connect((self.sdrplay3_rspduo, 0), (self.spectre_batched_file_sink, 0))


class tuner_1_swept_center_frequency(spectre_top_block):
    def flowgraph(self, tag: str, parameters: Parameters) -> None:

        # Unpack the capture config parameters
        sample_rate = parameters.get_parameter_value(PName.SAMPLE_RATE)
        bandwidth = parameters.get_parameter_value(PName.BANDWIDTH)
        min_frequency = parameters.get_parameter_value(PName.MIN_FREQUENCY)
        max_frequency = parameters.get_parameter_value(PName.MAX_FREQUENCY)
        frequency_step = parameters.get_parameter_value(PName.FREQUENCY_STEP)
        samples_per_step = parameters.get_parameter_value(PName.SAMPLES_PER_STEP)
        if_gain = parameters.get_parameter_value(PName.IF_GAIN)
        rf_gain = parameters.get_parameter_value(PName.RF_GAIN)
        batch_size = parameters.get_parameter_value(PName.BATCH_SIZE)

        # Blocks
        self.spectre_sweep_driver = spectre.sweep_driver(
            min_frequency,
            max_frequency,
            frequency_step,
            sample_rate,
            samples_per_step,
            "freq",
        )
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            get_batches_dir_path(),
            tag,
            batch_size,
            sample_rate,
            True,
            "freq",
            min_frequency,
        )
        self.sdrplay3_rspduo = sdrplay3.rspduo(
            "",
            rspduo_mode="Single Tuner",
            antenna="Tuner 1 50 ohm",
            stream_args=sdrplay3.stream_args(output_type="fc32", channels_size=1),
        )
        self.sdrplay3_rspduo.set_sample_rate(sample_rate, True)
        self.sdrplay3_rspduo.set_center_freq(min_frequency, True)
        self.sdrplay3_rspduo.set_bandwidth(bandwidth)
        self.sdrplay3_rspduo.set_antenna("Tuner 1 50 ohm")
        self.sdrplay3_rspduo.set_gain_mode(False)
        self.sdrplay3_rspduo.set_gain(if_gain, "IF", True)
        self.sdrplay3_rspduo.set_gain(rf_gain, "RF", True)
        self.sdrplay3_rspduo.set_freq_corr(0)
        self.sdrplay3_rspduo.set_dc_offset_mode(False)
        self.sdrplay3_rspduo.set_iq_balance_mode(False)
        self.sdrplay3_rspduo.set_agc_setpoint(-30)
        self.sdrplay3_rspduo.set_rf_notch_filter(False)
        self.sdrplay3_rspduo.set_dab_notch_filter(True)
        self.sdrplay3_rspduo.set_am_notch_filter(False)
        self.sdrplay3_rspduo.set_biasT(False)
        self.sdrplay3_rspduo.set_stream_tags(True)
        self.sdrplay3_rspduo.set_debug_mode(False)
        self.sdrplay3_rspduo.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo.set_show_gain_changes(False)

        # Connections
        self.msg_connect(
            (self.spectre_sweep_driver, "retune_command"),
            (self.sdrplay3_rspduo, "command"),
        )
        self.connect((self.sdrplay3_rspduo, 0), (self.spectre_batched_file_sink, 0))
        self.connect((self.sdrplay3_rspduo, 0), (self.spectre_sweep_driver, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from typing import Callable, Optional, cast
from functools import partial

from spectre_core.capture_configs import (
    CaptureTemplate,
    CaptureMode,
    Parameters,
    Bound,
    PName,
    get_base_capture_template,
    make_base_capture_template,
    get_base_ptemplate,
    validate_window,
)
from ._signal_generator_gr import cosine_wave, constant_staircase
from ._gr import capture
from ._receiver_names import ReceiverName
from .._receiver import Receiver
from .._specs import SpecName
from .._register import register_receiver


def _make_capture_template_cos_wave(receiver: Receiver) -> CaptureTemplate:
    capture_template = get_base_capture_template(CaptureMode.FIXED_CENTER_FREQUENCY)
    capture_template.add_ptemplate(get_base_ptemplate(PName.AMPLITUDE))
    capture_template.add_ptemplate(get_base_ptemplate(PName.FREQUENCY))

    capture_template.set_defaults(
        (PName.BATCH_SIZE, 3.0),
        (PName.CENTER_FREQUENCY, 16000),
        (PName.AMPLITUDE, 2.0),
        (PName.FREQUENCY, 32000),
        (PName.SAMPLE_RATE, 128000),
        (PName.WINDOW_HOP, 512),
        (PName.WINDOW_SIZE, 512),
        (PName.WINDOW_TYPE, "boxcar"),
    )

    capture_template.enforce_defaults(
        PName.TIME_RESOLUTION,
        PName.TIME_RANGE,
        PName.FREQUENCY_RESOLUTION,
        PName.WINDOW_TYPE,
    )

    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.FREQUENCY,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND),
            )
        ],
    )
    return capture_template


def _make_capture_template_constant_staircase(receiver: Receiver) -> CaptureTemplate:
    capture_template = make_base_capture_template(
        PName.TIME_RESOLUTION,
        PName.FREQUENCY_RESOLUTION,
        PName.TIME_RANGE,
        PName.SAMPLE_RATE,
        PName.BATCH_SIZE,
        PName.WINDOW_TYPE,
        PName.WINDOW_HOP,
        PName.WINDOW_SIZE,
        PName.EVENT_HANDLER_KEY,
        PName.BATCH_KEY,
        PName.WATCH_EXTENSION,
        PName.MIN_SAMPLES_PER_STEP,
        PName.MAX_SAMPLES_PER_STEP,
        PName.FREQUENCY_STEP,
        PName.STEP_INCREMENT,
        PName.OBS_ALT,
        PName.OBS_LAT,
        PName.OBS_LON,
        PName.OBJECT,
        PName.ORIGIN,
        PName.TELESCOPE,
        PName.INSTRUMENT,
    )

    capture_template.set_defaults(
        (PName.BATCH_SIZE, 3.0),
        (PName.FREQUENCY_STEP, 128000),
        (PName.MAX_SAMPLES_PER_STEP, 5000),
        (PName.MIN_SAMPLES_PER_STEP, 4000),
        (PName.SAMPLE_RATE, 128000),
        (PName.STEP_INCREMENT, 200),
        (PName.WINDOW_HOP, 512),
        (PName.WINDOW_SIZE, 512),
        (PName.WINDOW_TYPE, "boxcar"),
        (PName.EVENT_HANDLER_KEY, "swept_center_frequency"),
        (PName.BATCH_KEY, "iq_stream"),
        (PName.WATCH_EXTENSION, "bin"),
    )

    capture_template.enforce_defaults(
        PName.TIME_RESOLUTION,
        PName.TIME_RANGE,
        PName.FREQUENCY_RESOLUTION,
        PName.WINDOW_TYPE,
        PName.EVENT_HANDLER_KEY,
        PName.BATCH_KEY,
        PName.WATCH_EXTENSION,
    )

    return capture_template


def _make_pvalidator_cosine_wave(receiver: Receiver) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_window(parameters)

        sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
        window_size = cast(int, parameters.get_parameter_value(PName.WINDOW_SIZE))
        frequency = cast(float, parameters.get_parameter_value(PName.FREQUENCY))

        # check that the sample rate is an integer multiple of the underlying signal frequency
        if sample_rate % frequency != 0:
            raise ValueError(
                "The sampling rate must be some integer multiple of frequency"
            )

        a = sample_rate / frequency
        if a < 2:
            raise ValueError(
                (
                    f"The ratio of sampling rate over frequency must be greater than two. "
                    f"Got {a}"
                )
            )

        # analytical requirement
        # if p is the number of sampled cycles, we can find that p = window_size / a
        # the number of sampled cycles must be a positive natural number.
        p = window_size / a
        if window_size % a != 0:
            raise ValueError(
                (
                    f"The number of sampled cycles must be a positive natural number. "
                    f"Computed that p={p}"
                )
            )

    return pvalidator


def _make_pvalidator_constant_staircase(
    receiver: Receiver,
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_window(parameters)

        freq_step = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))
        sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
        min_samples_per_step = cast(
            int, parameters.get_parameter_value(PName.MIN_SAMPLES_PER_STEP)
        )
        max_samples_per_step = cast(
            int, parameters.get_parameter_value(PName.MAX_SAMPLES_PER_STEP)
        )

        if freq_step != sample_rate:
            raise ValueError(f"The frequency step must be equal to the sampling rate")

        if min_samples_per_step > max_samples_per_step:
            raise ValueError(
                (
                    f"Minimum samples per step cannot be greater than the maximum samples per step. "
                    f"Got {min_samples_per_step}, which is greater than {max_samples_per_step}"
                )
            )

    return pvalidator


@dataclass(frozen=True)
class _Mode:
    """An operating mode for the `SignalGenerator` receiver."""

    COSINE_WAVE = "cosine_wave"
    CONSTANT_STAIRCASE = "constant_staircase"


@register_receiver(ReceiverName.SIGNAL_GENERATOR)
class SignalGenerator(Receiver):
    """An entirely software-defined receiver, which generates synthetic signals."""

    def __init__(self, name: ReceiverName, mode: Optional[str] = None) -> None:
        super().__init__(name, mode)

        self.add_spec(SpecName.SAMPLE_RATE_LOWER_BOUND, 64000)
        self.add_spec(SpecName.SAMPLE_RATE_UPPER_BOUND, 640000)
        self.add_spec(SpecName.FREQUENCY_LOWER_BOUND, 16000)
        self.add_spec(SpecName.FREQUENCY_UPPER_BOUND, 160000)

        self.add_mode(
            _Mode.COSINE_WAVE,
            partial(capture, top_block_cls=cosine_wave),
            _make_capture_template_cos_wave(self),
            _make_pvalidator_cosine_wave(self),
        )

        self.add_mode(
            _Mode.CONSTANT_STAIRCASE,
            partial(capture, top_block_cls=constant_staircase),
            _make_capture_template_constant_staircase(self),
            _make_pvalidator_constant_staircase(self),
        )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from copy import deepcopy
from typing import Any, Iterator

from ._capture_modes import CaptureMode
from ._parameters import Parameter, Parameters
from ._pconstraints import BasePConstraint
from ._ptemplates import PTemplate, get_base_ptemplate
from ._ptemplates import PName


class CaptureTemplate:
    """A managed collection of parameter templates. Strictly defines what parameters are required
    in a capture config, and the values each parameter can take.
    """

    def __init__(self) -> None:
        """Initialise a `CaptureTemplate` instance."""
        self._ptemplates: dict[PName, PTemplate] = {}

    @property
    def name_list(self) -> list[PName]:
        """The names of all required parameters in the capture template."""
        return list(self._ptemplates.keys())

    def add_ptemplate(self, ptemplate: PTemplate) -> None:
        """Add a parameter template to the capture template.

        :param ptemplate: Describes a required parameter for this capture template.
        """
        self._ptemplates[ptemplate.name] = ptemplate

    def get_ptemplate(self, parameter_name: PName) -> PTemplate:
        """Get the parameter template corresponding to the parameter with the name `parameter_name`.

        :param parameter_name: The name of the parameter.
        :return: The corresponding `PTemplate` instance.
        :raises ValueError: If the parameter name is not found in the template.
        """
        if parameter_name not in self._ptemplates:
            raise ValueError(
                f"Parameter with name '{parameter_name}' is not found in the template. "
                f"Expected one of {self.name_list}"
            )
        return self._ptemplates[parameter_name]

    def __apply_parameter_template(
        self, parameter: Parameter, apply_pconstraints: bool = True
    ) -> None:
        """Apply the corresponding parameter template to the input parameter.

        As a side effect, the value of the input parameter will be type cast
        according to the template.
        """
        ptemplate = self.get_ptemplate(parameter.name)
        parameter.value = ptemplate.apply_template(parameter.value, apply_pconstraints)

    def __apply_parameter_templates(
        self, parameters: Parameters, apply_pconstraints: bool = True
    ) -> None:
        """Apply the corresponding parameter template to each of the input parameters."""
        for parameter in parameters:
            self.__apply_parameter_template(parameter, apply_pconstraints)

    def __fill_missing_with_defaults(self, parameters: Parameters) -> None:
        """Add default parameters to `parameters` for any missing entries.

        Missing parameters are identified by comparing `parameters` against the
        current capture template. Defaults are derived from the corresponding
        parameter templates.
        """
        for ptemplate in self:
            if ptemplate.name not in parameters.name_list:
                # no args for `make_parameter` implies the parameter with the default value will be used.
                parameter = ptemplate.make_parameter()
                parameters.add_parameter(parameter.name, parameter.value)

    def apply_template(
        self, parameters: Parameters, apply_pconstraints: bool = True
    ) -> Parameters:
        """Apply the capture template to the input parameters. This involves:

        - Adding default parameters if they are missing with respect to this template.
        - Type casting the value of each input parameter according to the corresponding parameter template.
        - Validating the value of each input parameter against any corresponding pconstraints.

        :param parameters: The parameters to apply this capture template to.
        :param apply_pconstraints: If True, apply `PConstraints` to each of the input parameters.
        :return: A `Parameters` instance compliant with this template.
        """
        self.__fill_missing_with_defaults(parameters)
        self.__apply_parameter_templates(parameters, apply_pconstraints)
        return parameters

    def __iter__(self) -> Iterator[PTemplate]:
        """Iterate over stored ptemplates."""
        yield from self._ptemplates.values()

    def set_default(self, parameter_name: PName, default: Any) -> None:
        """Set the default of an existing parameter template.

        :param parameter_name: The name of the parameter template to be updated.
        :param default: The new default value.
        """
        self.get_ptemplate(parameter_name).default = default

    def set_defaults(self, *ptuples: tuple[PName, Any]) -> None:
        """Update the defaults of multiple parameter templates.

        :param ptuples: Tuples of the form (`parameter_name`, `new_default`) to update defaults.
        """
        for parameter_name, default in ptuples:
            self.set_default(parameter_name, default)

    def enforce_default(self, parameter_name: PName) -> None:
        """Set the `enforce_default` attribute of an existing parameter template to True.

        :param parameter_name: The name of the parameter template to enforce its default value.
        """
        self.get_ptemplate(parameter_name).enforce_default = True

    def enforce_defaults(self, *parameter_names: PName) -> None:
        """Set the `enforce_default` attribute of multiple existing parameter templates to True.

        :param parameter_names: The names of the parameter templates to enforce their default values.
        """
        for name in parameter_names:
            self.enforce_default(name)

    def add_pconstraint(
        self, parameter_name: PName, pconstraints: list[BasePConstraint]
    ) -> None:
        """Add one or more `PConstraint` instances to an existing parameter template.

        :param parameter_name: The name of the parameter template to add constraints to.
        :param pconstraints: A list of `PConstraint` instances to be added.
        """
        for pconstraint in pconstraints:
            self.get_ptemplate(parameter_name).add_pconstraint(pconstraint)

    def to_dict(self) -> dict[str, dict[str, str]]:
        """Convert the current instance to a serialisable dictionary.

        :return: A dictionary representation of this capture template, where all values
        are formatted strings.
        """
        return {ptemplate.name.value: ptemplate.to_dict() for ptemplate in self}


def make_base_capture_template(*pnames: PName) -> CaptureTemplate:
    """Make a capture template composed entirely of base `PTemplate` instances.

    :param pnames: The names of parameters to include in the capture template.
    :return: A capture template composed of base parameter templates.
    """
    capture_template = CaptureTemplate()
    for pname in pnames:
        capture_template.add_ptemplate(get_base_ptemplate(pname))
    return capture_template


def _make_fixed_frequency_capture_template() -> CaptureTemplate:
    """The absolute minimum required parameters for any fixed frequency capture template."""
    capture_template = make_base_capture_template(
        PName.BATCH_SIZE,
        PName.CENTER_FREQUENCY,
        PName.BATCH_KEY,
        PName.EVENT_HANDLER_KEY,
        PName.FREQUENCY_RESOLUTION,
        PName.INSTRUMENT,
        PName.OBS_ALT,
        PName.OBS_LAT,
        PName.OBS_LON,
        PName.OBJECT,
        PName.ORIGIN,
        PName.SAMPLE_RATE,
        PName.TELESCOPE,
        PName.TIME_RANGE,
        PName.TIME_RESOLUTION,
        PName.WATCH_EXTENSION,
        PName.WINDOW_HOP,
        PName.WINDOW_SIZE,
        PName.WINDOW_TYPE,
    )
    capture_template.set_defaults(
        (PName.EVENT_HANDLER_KEY, "fixed_center_frequency"),
        (PName.BATCH_KEY, "iq_stream"),
        (PName.WATCH_EXTENSION, "bin"),
    )
    capture_template.enforce_defaults(
        PName.EVENT_HANDLER_KEY, PName.BATCH_KEY, PName.WATCH_EXTENSION
    )
    return capture_template


def _make_swept_frequency_capture_template() -> CaptureTemplate:
    """The absolute minimum required parameters for any swept frequency capture template."""
    capture_template = make_base_capture_template(
        PName.BATCH_SIZE,
        PName.BATCH_KEY,
        PName.EVENT_HANDLER_KEY,
        PName.FREQUENCY_RESOLUTION,
        PName.FREQUENCY_STEP,
        PName.INSTRUMENT,
        PName.MAX_FREQUENCY,
        PName.MIN_FREQUENCY,
        PName.OBS_ALT,
        PName.OBS_LAT,
        PName.OBS_LON,
        PName.OBJECT,
        PName.ORIGIN,
        PName.SAMPLE_RATE,
        PName.SAMPLES_PER_STEP,
        PName.TELESCOPE,
        PName.TIME_RANGE,
        PName.TIME_RESOLUTION,
        PName.WATCH_EXTENSION,
        PName.WINDOW_HOP,
        PName.WINDOW_SIZE,
        PName.WINDOW_TYPE,
    )
    capture_template.set_defaults(
        (PName.EVENT_HANDLER_KEY, "swept_center_frequency"),
        (PName.BATCH_KEY, "iq_stream"),
        (PName.WATCH_EXTENSION, "bin"),
    )
    capture_template.enforce_defaults(
        PName.EVENT_HANDLER_KEY, PName.BATCH_KEY, PName.WATCH_EXTENSION
    )
    return capture_template


_base_capture_templates: dict[CaptureMode, CaptureTemplate] = {
    CaptureMode.FIXED_CENTER_FREQUENCY: _make_fixed_frequency_capture_template(),
    CaptureMode.SWEPT_CENTER_FREQUENCY: _make_swept_frequency_capture_template(),
}


def get_base_capture_template(capture_mode: CaptureMode) -> CaptureTemplate:
    """Get a pre-defined capture template, to be configured according to the specific use-case.

    :param capture_mode: The mode used to retrieve the capture template.
    :return: A deep copy of the template for the specified mode.
    :raises KeyError: If no capture template is found for the given mode.
    """
    if capture_mode not in _base_capture_templates:
        raise KeyError(
            f"No capture template found for the capture mode '{capture_mode}'. "
            f"Expected one of {list(_base_capture_templates.keys())}"
        )
    return deepcopy(_base_capture_templates[capture_mode])



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
from time import sleep

from spectre_core.jobs import Worker, Job, make_worker, Duration


def _short_sleep() -> None:
    sleep(Duration.ONE_CENTISECOND)


def _sleep_forever() -> None:
    """Sleep indefinetely."""
    while True:
        _short_sleep()


def _fail_instantly() -> None:
    raise RuntimeError("Boom!")


def _make_successful_runtime_worker() -> Worker:
    return make_worker("successful_runtime_worker", _sleep_forever)


def _make_instantly_failing_runtime_worker() -> Worker:
    return make_worker("instantly_failing_runtime_worker", _fail_instantly)


@pytest.fixture
def successful_runtime_worker() -> Worker:
    """A worker which models successful runtime.

    The created `Worker` instance manages a process which sleeps indefinitely.
    """
    return _make_successful_runtime_worker()


@pytest.fixture
def instantly_failing_runtime_worker() -> Worker:
    """A worker which models runtime that fails instantly.

    The created `Worker` instance manages a process which instantly fails.
    """
    return _make_instantly_failing_runtime_worker()


@pytest.fixture
def successful_runtime_job() -> Job:
    """Create a job modelling successful runtime.

    Return a `Job` instance,  where each worker sleeps indefinitely.
    """
    _num_workers = 2  # arbitrarily choose two workers
    workers = [_make_successful_runtime_worker() for _ in range(_num_workers)]
    return Job(workers)


@pytest.fixture
def partially_failing_job() -> Job:
    """Create a job modelling partially failing runtime.

    Return a `Job` instance, where one worker sleeps indefinitely, and one fails instantly.
    """
    workers = [
        _make_instantly_failing_runtime_worker(),
        _make_successful_runtime_worker(),
    ]
    return Job(workers)


class TestWorker:
    def test_name(self, successful_runtime_worker: Worker) -> None:
        """Check that the name of the process is as expected."""
        assert successful_runtime_worker.name == "successful_runtime_worker"

    def test_is_alive(self, successful_runtime_worker: Worker) -> None:
        """Check that a worker which is successfully running is alive."""
        successful_runtime_worker.start()
        _short_sleep()
        assert successful_runtime_worker.is_alive

        # Kill the process, since otherwise they would keep running until the parent process terminates.
        successful_runtime_worker.kill()

    def test_is_not_alive(self, instantly_failing_runtime_worker: Worker) -> None:
        """Check that a worker which instantly failed, is not alive."""
        instantly_failing_runtime_worker.start()
        _short_sleep()
        assert not instantly_failing_runtime_worker.is_alive

    def test_kill(self, successful_runtime_worker: Worker) -> None:
        """Check that killing the worker, results in it not being alive."""
        successful_runtime_worker.start()
        successful_runtime_worker.kill()
        _short_sleep()
        assert not successful_runtime_worker.is_alive

    def test_restart(self, successful_runtime_worker: Worker) -> None:
        """Check that a restarted worker evaluates as alive."""
        successful_runtime_worker.start()
        successful_runtime_worker.restart()
        _short_sleep()
        assert successful_runtime_worker.is_alive

        # Kill the process, since otherwise they would keep running until the parent process terminates.
        successful_runtime_worker.kill()

    def test_multiple_restarts(self, successful_runtime_worker: Worker) -> None:
        """Check that a worker can be restarted multiple times."""
        successful_runtime_worker.start()
        successful_runtime_worker.restart()
        successful_runtime_worker.restart()

        _short_sleep()

        assert successful_runtime_worker.is_alive

        # Kill the process, since otherwise they would keep running until the parent process terminates.
        successful_runtime_worker.kill()

    def test_starting_twice(self, successful_runtime_worker: Worker) -> None:
        """Check that a worker cannot be started twice."""
        with pytest.raises(RuntimeError):
            successful_runtime_worker.start()
            successful_runtime_worker.start()

        # Kill the process, since otherwise they would keep running until the parent process terminates.
        successful_runtime_worker.kill()

    def test_killing_dead_process(self, successful_runtime_worker: Worker) -> None:
        """Check that we cannot kill a process which is not alive."""
        with pytest.raises(RuntimeError):
            successful_runtime_worker.kill()


class TestJobs:

    def test_start(self, successful_runtime_job: Job) -> None:
        """Check that when a job starts, the workers are all alive."""
        successful_runtime_job.start()
        _short_sleep()
        assert successful_runtime_job.workers_are_alive

        # Kill the workers, since otherwise they would keep running until the parent process terminates.
        successful_runtime_job.kill()

    def test_kill(self, successful_runtime_job: Job) -> None:
        """Check that when a job is started, then killed, that the workers are not alive."""
        successful_runtime_job.start()
        successful_runtime_job.kill()
        _short_sleep()

        # Check all the workers are not alive.
        assert not successful_runtime_job.workers_are_alive

    def test_monitor_successful_job(self, successful_runtime_job: Job) -> None:
        """Check that once the total runtime of a job is complete, the workers are no longer alive."""
        successful_runtime_job.start()
        successful_runtime_job.monitor(Duration.ONE_CENTISECOND)

        # Sleep for a moment, to give the job time to kill the workers once the total runtime has elapsed.
        _short_sleep()

        # Check all the workers are not alive.
        assert not successful_runtime_job.workers_are_alive

    def test_monitor_failed_job(self, partially_failing_job: Job) -> None:
        """Check that if a worker fails, and force restart is false, that the main process raises a `RuntimeError`."""
        partially_failing_job.start()
        with pytest.raises(RuntimeError):
            partially_failing_job.monitor(Duration.ONE_SECOND)

        # Check all the workers are not alive.
        assert not partially_failing_job.workers_are_alive

    def test_max_restarts(self, partially_failing_job: Job) -> None:
        """Check that we don't get any neverending force restart loops."""
        partially_failing_job.start()
        _max_restarts = 3
        with pytest.raises(
            RuntimeError,
            match=f"Maximum number of restarts has been reached: {_max_restarts}",
        ):
            partially_failing_job.monitor(
                Duration.TEN_SECONDS, force_restart=True, max_restarts=_max_restarts
            )

        # Check all the workers are not alive.
        assert not partially_failing_job.workers_are_alive

    def test_single_restart(self, successful_runtime_job: Job) -> None:
        """Check that we can restart a job multiple once, and the workers are alive afterwards."""
        successful_runtime_job.start()
        successful_runtime_job.restart()
        _short_sleep()
        assert successful_runtime_job.workers_are_alive

        # Kill the workers, since otherwise they would keep running until the parent process terminates.
        successful_runtime_job.kill()

    def test_multiple_restarts(self, successful_runtime_job: Job) -> None:
        """Check that we can restart a job multiple times, and that the workers are alive afterwards."""
        successful_runtime_job.start()
        successful_runtime_job.restart()
        successful_runtime_job.restart()
        _short_sleep()
        assert successful_runtime_job.workers_are_alive

        # Kill the workers, since otherwise they would keep running until the parent process terminates.
        successful_runtime_job.kill()

    def test_starting_twice(self, successful_runtime_job: Job) -> None:
        """Check that we cannot start a job twice."""
        with pytest.raises(RuntimeError):
            successful_runtime_job.start()
            successful_runtime_job.start()

        # Kill the workers, since otherwise they would keep running until the parent process terminates.
        successful_runtime_job.kill()



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from dataclasses import dataclass
from typing import Optional

import numpy as np
import numpy.typing as npt
from astropy.io import fits
from astropy.io.fits.hdu.image import PrimaryHDU
from astropy.io.fits.hdu.table import BinTableHDU
from astropy.io.fits.hdu.hdulist import HDUList

from spectre_core.exceptions import InvalidSweepMetadataError
from spectre_core.config import TimeFormat
from spectre_core.spectrograms import Spectrogram, SpectrumUnit
from ._batch_keys import BatchKey
from .._base import BaseBatch, BatchFile
from .._register import register_batch


@dataclass(frozen=True)
class _BatchExtension:
    """Supported extensions for a `IQStreamBatch`.

    :ivar FITS: Corresponds to the `.fits` file extension.
    :ivar BIN: Corresponds to the `.bin` file extension.
    :ivar HDR: Corresponds to the `.hdr` file extension.
    """

    FITS: str = "fits"
    BIN: str = "bin"
    HDR: str = "hdr"


class _BinFile(BatchFile[npt.NDArray[np.complex64]]):
    """Stores complex IQ samples in the binary format, as produced by the `gr-spectre`
    OOT module block `batched_file_sink`.
    """

    def __init__(self, batch_parent_dir_path: str, batch_name: str) -> None:
        """Initialise a `_BinFile` instance.

        :param batch_parent_dir_path: The parent directory for the batch.
        :param batch_name: The batch name.
        """
        super().__init__(batch_parent_dir_path, batch_name, _BatchExtension.BIN)

    def _read(self) -> npt.NDArray[np.complex64]:
        """Reads the binary file and returns the stored complex IQ samples.

        :return: The raw 32-bit floats in the binary file, interpreted as 64-bit complex IQ samples.
        """
        with open(self.file_path, "rb") as fh:
            return np.fromfile(fh, dtype=np.complex64)


@dataclass
class IQMetadata:
    """Represents metadata for IQ samples produced by the `gr-spectre` OOT module block `batched_file_sink`.

    :ivar millisecond_correction: The millisecond component of the batch start time.
    :ivar center_frequencies: Center frequencies for each IQ sample, if the stream was frequency tagged.
    None otherwise.
    :ivar num_samples: Number of samples collected at each center frequency, if the stream was frequency
    tagged. None otherwise.
    """

    millisecond_correction: int
    center_frequencies: Optional[npt.NDArray[np.float32]] = None
    num_samples: Optional[npt.NDArray[np.int32]] = None

    def is_frequency_tagged(self) -> bool:
        """Check if the IQ metadata contains frequency tagging information.

        :return: True if frequency tagging information is present; False otherwise.
        """
        return (self.center_frequencies is not None) and (self.num_samples is not None)


class _HdrFile(BatchFile[IQMetadata]):
    """Stores IQ sample metadata produced by the `gr-spectre` OOT module block `batched_file_sink`, used
    to help parse the corresponding `.bin` file.

    File Structure:
        - If frequency tagged:
            (`<millisecond_correction>`, `<freq_0>`, `<num_samples_0>`, `<freq_1>`, `<num_samples_1>`, ...)
            All values are stored as 32-bit floats.
            - The first value is the millisecond component for the batch start time.
            - Subsequent tuples (`<freq_n>`, `<num_samples_n>`) indicate that `<num_samples_n>` samples were collected at `<freq_n>`.
        - If not frequency tagged:
            (`<millisecond_correction>`)
            Only the millisecond correction is present, with no frequency information.

    This format enables mapping IQ samples in the binary file to their corresponding center frequencies, if applicable.
    """

    def __init__(self, parent_dir_path: str, base_file_name: str) -> None:
        """Initialise a `_HdrFile` instance.

        :param parent_dir_path: The parent directory for the batch.
        :param base_file_name: The batch name.
        """
        super().__init__(parent_dir_path, base_file_name, _BatchExtension.HDR)

    def _read(self) -> IQMetadata:
        """Parses the binary contents of the `.hdr` file to extract IQ sample metadata.

        :return: An instance of `IQMetadata` containing the parsed metadata, including the millisecond correction
        and, if applicable, frequency tagging details.
        """
        hdr_contents = self._extract_raw_contents()
        millisecond_correction = self._get_millisecond_correction(hdr_contents)
        if hdr_contents.size == 1:
            return IQMetadata(millisecond_correction)
        else:
            center_frequencies = self._get_center_frequencies(hdr_contents)
            num_samples = self._get_num_samples(hdr_contents)
            self._validate_frequencies_and_samples(center_frequencies, num_samples)
            return IQMetadata(millisecond_correction, center_frequencies, num_samples)

    def _extract_raw_contents(self) -> npt.NDArray[np.float32]:
        """Reads the raw contents of the `.hdr` file."""
        with open(self.file_path, "rb") as fh:
            # the `batched_file_sink` block in the `gr-spectre` GNU Radio OOT module stores
            # the integral millisecond component as a 32-bit float.
            return np.fromfile(fh, dtype=np.float32)

    def _get_millisecond_correction(self, hdr_contents: npt.NDArray[np.float32]) -> int:
        """Extracts and validates the millisecond component of the batch start time.

        The value is stored as a 32-bit float but interpreted as an integer.
        """
        millisecond_correction = float(hdr_contents[0])

        if not millisecond_correction.is_integer():
            raise TypeError(
                f"Expected integer value for millisecond correction, but got {millisecond_correction}"
            )

        return int(millisecond_correction)

    def _get_center_frequencies(
        self, hdr_contents: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Extracts the center frequencies from the `.hdr` file contents.

        Center frequencies are stored at every second entry, starting from the first index.
        """
        return hdr_contents[1::2]

    def _get_num_samples(
        self, hdr_contents: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.int32]:
        """Extracts the number of samples per frequency from the `.hdr` file contents.

        The values are stored as 32-bit floats but are interpreted as integers.
        Sample counts are located at every second entry, starting from the second index.
        """
        num_samples_as_float = hdr_contents[2::2]
        if not all(num_samples_as_float == num_samples_as_float.astype(int)):
            raise InvalidSweepMetadataError(
                "Number of samples per frequency is expected to describe an integer"
            )
        return num_samples_as_float.astype(np.int32)

    def _validate_frequencies_and_samples(
        self,
        center_frequencies: npt.NDArray[np.float32],
        num_samples: npt.NDArray[np.int32],
    ) -> None:
        """Ensures that each center frequency has a corresponding sample count."""
        if len(center_frequencies) != len(num_samples):
            raise InvalidSweepMetadataError(
                "Center frequencies and number of samples arrays are not the same length"
            )


class _FitsFile(BatchFile[Spectrogram]):
    """Stores spectrogram data in the FITS file format, as generated by `spectre` from a stream of IQ samples."""

    def __init__(self, parent_dir_path: str, base_file_name: str) -> None:
        """Initialise a `_FitsFile` instance.

        :param parent_dir_path: The parent directory for the batch.
        :param base_file_name: The batch name.
        """
        super().__init__(parent_dir_path, base_file_name, _BatchExtension.FITS)

    def _read(self) -> Spectrogram:
        """Read the FITS file and create a spectrogram.

        :return: A `Spectrogram` instance containing the parsed FITS file data.
        """
        with fits.open(self.file_path, mode="readonly") as hdulist:
            primary_hdu = self._get_primary_hdu(hdulist)
            dynamic_spectra = self._get_dynamic_spectra(primary_hdu)
            bunit = self._get_bunit(primary_hdu)
            spectrogram_start_datetime = self._get_spectrogram_start_datetime(
                primary_hdu
            )
            bintable_hdu = self._get_bintable_hdu(hdulist)
            times = self._get_times(bintable_hdu)
            frequencies = self._get_frequencies(bintable_hdu)

        # bunit is interpreted as a SpectrumUnit.
        spectrum_unit = SpectrumUnit(bunit)
        return Spectrogram(
            dynamic_spectra,
            times,
            frequencies,
            self.tag,
            spectrum_unit,
            spectrogram_start_datetime,
        )

    def _get_primary_hdu(self, hdulist: HDUList) -> PrimaryHDU:
        return hdulist["PRIMARY"]

    def _get_bintable_hdu(self, hdulist: HDUList) -> BinTableHDU:
        return hdulist[1]

    def _get_bunit(self, primary_hdu: PrimaryHDU) -> str:
        """Get the units corresponding to the elements of the dynamic spectra."""
        return primary_hdu.header["BUNIT"]

    def _get_dynamic_spectra(self, primary_hdu: PrimaryHDU) -> npt.NDArray[np.float32]:
        return primary_hdu.data

    def _get_spectrogram_start_datetime(self, primary_hdu: PrimaryHDU) -> datetime:
        """Get the start time of the spectrogram, up to the full precision available."""
        date_obs = primary_hdu.header["DATE-OBS"]
        time_obs = primary_hdu.header["TIME-OBS"]
        return datetime.strptime(f"{date_obs}T{time_obs}", TimeFormat.PRECISE_DATETIME)

    def _get_times(self, bintable_hdu: BinTableHDU) -> npt.NDArray[np.float32]:
        """Extracts the elapsed times for each spectrum in seconds, with the first spectrum set to t=0
        by convention.
        """
        return bintable_hdu.data["TIME"][0]  # already in seconds

    def _get_frequencies(self, bintable_hdu: BinTableHDU) -> npt.NDArray[np.float32]:
        """Extracts the frequencies for each spectral component."""
        frequencies_MHz = bintable_hdu.data["FREQUENCY"][0]
        return frequencies_MHz * 1e6  # convert to Hz


@register_batch(BatchKey.IQ_STREAM)
class IQStreamBatch(BaseBatch):
    """A batch of data derived from a stream of IQ samples from some receiver.

    Supports the following extensions:
    - `.fits` (via the `spectrogram_file` attribute)
    - `.bin` (via the `bin_file` attribute)
    - `.hdr` (via the `hdr_file` attribute)
    """

    def __init__(self, start_time: str, tag: str) -> None:
        """Initialise a `IQStreamBatch` instance.

        :param start_time: The start time of the batch.
        :param tag: The batch name tag.
        """
        super().__init__(start_time, tag)
        self._fits_file = _FitsFile(self.parent_dir_path, self.name)
        self._bin_file = _BinFile(self.parent_dir_path, self.name)
        self._hdr_file = _HdrFile(self.parent_dir_path, self.name)

        # add files formally to the batch
        self.add_file(self.spectrogram_file)
        self.add_file(self.bin_file)
        self.add_file(self.hdr_file)

    @property
    def spectrogram_file(self) -> _FitsFile:
        """The batch file corresponding to the `.fits` extension."""
        return self._fits_file

    @property
    def bin_file(self) -> _BinFile:
        """The batch file corresponding to the `.bin` extension."""
        return self._bin_file

    @property
    def hdr_file(self) -> _HdrFile:
        """The batch file corresponding to the `.hdr` extension."""
        return self._hdr_file



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
import os
from spectre_core.config import (
    get_batches_dir_path,
    get_logs_dir_path,
    get_configs_dir_path,
)


@pytest.fixture(autouse=True)
def patch_spectre_data_dir_path():
    """Patch the environment which defines the shared ancestral path of all log, batch and config files."""
    pytest.MonkeyPatch().setenv(
        "SPECTRE_DATA_DIR_PATH", os.path.join("/tmp", ".spectre-data")
    )


@pytest.mark.parametrize(
    ["year", "month", "day", "expected_dir_path"],
    [
        (None, None, None, os.path.join("/tmp", ".spectre-data", "batches")),
        (2025, None, None, os.path.join("/tmp", ".spectre-data", "batches", "2025")),
        (2025, 2, None, os.path.join("/tmp", ".spectre-data", "batches", "2025", "02")),
        (
            2025,
            2,
            13,
            os.path.join("/tmp", ".spectre-data", "batches", "2025", "02", "13"),
        ),
    ],
)
def test_get_batches_dir_path(
    year: int,
    month: int,
    day: int,
    expected_dir_path: str,
) -> None:
    """Check that the batches directory paths are created as expected."""
    result = get_batches_dir_path(year, month, day)
    assert result == expected_dir_path


@pytest.mark.parametrize(
    ["year", "month", "day", "expected_dir_path"],
    [
        (None, None, None, os.path.join("/tmp", ".spectre-data", "logs")),
        (2025, None, None, os.path.join("/tmp", ".spectre-data", "logs", "2025")),
        (2025, 2, None, os.path.join("/tmp", ".spectre-data", "logs", "2025", "02")),
        (
            2025,
            2,
            13,
            os.path.join("/tmp", ".spectre-data", "logs", "2025", "02", "13"),
        ),
    ],
)
def test_get_logs_dir_path(
    year: int,
    month: int,
    day: int,
    expected_dir_path: str,
) -> None:
    """Check that the logs directory paths are created as expected."""
    result = get_logs_dir_path(year, month, day)
    assert result == expected_dir_path


def test_get_configs_dir_path():
    """Check that the configs directory path is created as expected."""
    assert get_configs_dir_path() == os.path.join("/tmp", ".spectre-data", "configs")



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import signal
from typing import Type

from gnuradio import gr

from spectre_core.capture_configs import Parameters


class spectre_top_block(gr.top_block):
    """A thin wrapper around GNU Radio's `gr.top_block` class.

    Propagate through the input tag and parameters to `flowgraph` in the constructor,
    to make the GNU Radio flow graphs easily configurable.
    """

    def flowgraph(self, tag: str, parameters: Parameters) -> None:
        """Define the flowgraph for the block.

        This method uses inline imports to allow the `spectre_core.receivers`
        module to work without requiring all Out-Of-Tree (OOT) modules to be installed.
        Only import the necessary OOT modules when implementing this method.
        """
        raise NotImplementedError()

    def __init__(self, tag: str, parameters: Parameters) -> None:
        """Create an instance of `spectre_top_block`.

        :param tag: The capture config tag.
        :param parameters: The capture config parameters
        """
        gr.top_block.__init__(self)
        self.flowgraph(tag, parameters)


def capture(
    tag: str,
    parameters: Parameters,
    top_block_cls: Type[spectre_top_block],
    max_noutput_items: int = 10000000,
) -> None:
    """
    Run a GNU Radio flowgraph with the given number of output items.

    Typically, this should be used with `partial` from `functools` to create
    a capture method. For example,

    .. code-block:: python
        capture_method = partial(capture, top_block_cls=<your GNU Radio top block>)

    :param tag: The capture config tag
    :param parameters: The parameters stored in the capture config
    :param top_block_cls: The subclass of `spectre_top_block`, defining the GNURadio flowgraph.
    :param max_noutput_items: The number of max `noutput_items` in the flowgraph. This controls
    the maximum number of output items any block will handle during a call to work. Defaults to 10000000
    (which has been chosen as per GNU Radio source code).
    """
    tb = top_block_cls(tag, parameters)

    def sig_handler(sig=None, frame=None):
        tb.stop()
        tb.wait()
        sys.exit(0)

    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    tb.run(max_noutput_items)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger

_LOGGER = getLogger(__name__)

import os
from typing import Optional, Iterator, Tuple
from collections import OrderedDict
from datetime import datetime

from spectre_core._file_io import TextHandler
from spectre_core.config import get_logs_dir_path, TimeFormat
from ._process_types import ProcessType


def parse_log_file_name(file_name: str) -> Tuple[str, str, str]:
    """Parse the file name of a log into a start time, process ID and process type."""
    base_file_name, _ = os.path.splitext(file_name)
    log_start_time, pid, process_type = base_file_name.split("_")
    return log_start_time, pid, process_type


class Log(TextHandler):
    """Interface to read log files generated by `spectre`."""

    def __init__(self, start_time: str, pid: str, process_type: ProcessType) -> None:
        """Initialise a `Log` instance.

        :param start_time: The timestamp when the log file was created.
        :param pid: The ID of the process writing to the log file.
        :param process_type: Indicates the type of process, as defined by `ProcessType`.
        """
        self._start_time = start_time
        self._pid = pid
        self._process_type = process_type.value

        dt = datetime.strptime(start_time, TimeFormat.DATETIME)
        parent_path = get_logs_dir_path(dt.year, dt.month, dt.day)
        base_file_name = f"{start_time}_{pid}_{process_type.value}"

        super().__init__(parent_path, base_file_name, "log")

    @property
    def start_time(self) -> str:
        """The system time when the log was created."""
        return self._start_time

    @property
    def pid(self) -> str:
        """The ID of the process writing to the log file."""
        return self._pid

    @property
    def process_type(self) -> str:
        """Indicates the type of process, as defined by `ProcessType`."""
        return self._process_type


class Logs:
    """Filter and read a collection of logs generated by `spectre`."""

    def __init__(
        self,
        process_type: Optional[ProcessType] = None,
        year: Optional[int] = None,
        month: Optional[int] = None,
        day: Optional[int] = None,
    ) -> None:
        """Initialise a `Logs` instance.

        :param process_type: Filter by the process type. Defaults to None.
        :param year: Filter by the numeric year. Defaults to None.
        :param month: Filter by the numeric month. Defaults to None.
        :param day: Filter by the numeric day. Defaults to None.
        """
        self._process_type = process_type.value if process_type is not None else None

        self._log_map: dict[str, Log] = OrderedDict()
        self.set_date(year, month, day)

    @property
    def process_type(self) -> Optional[str]:
        """Indicates the type of process, as defined by `ProcessType`."""
        return self._process_type

    @property
    def year(self) -> Optional[int]:
        """Filter by the numeric year."""
        return self._year

    @property
    def month(self) -> Optional[int]:
        """Filter by the numeric month."""
        return self._month

    @property
    def day(self) -> Optional[int]:
        """Filter by the numeric day."""
        return self._day

    @property
    def logs_dir_path(self) -> str:
        """The shared ancestral path for all the log files. `Logs` recursively searches
        this directory to find all log files according to the date and process type."""
        return get_logs_dir_path(self.year, self.month, self.day)

    @property
    def log_list(self) -> list[Log]:
        """A list of all log handlers representing files found within `logs_dir_path`."""
        return list(self._log_map.values())

    @property
    def num_logs(self) -> int:
        """The number of log files found within `logs_dir_path`."""
        return len(self.log_list)

    @property
    def file_names(self) -> list[str]:
        """A list of all log file names found within `logs_dir_path`."""
        return list(self._log_map.keys())

    def set_date(
        self, year: Optional[int], month: Optional[int], day: Optional[int]
    ) -> None:
        """Reset `logs_dir_path` according to the numeric date, and refresh the list
        of available log files.

        :param year: The numeric year.
        :param month: The numeric month of the year.
        :param day: The numeric day of the month.
        """
        self._year = year
        self._month = month
        self._day = day
        self.update()

    def update(self) -> None:
        """Perform a fresh search of all files in `logs_dir_path` for log files
        according to the date and process type."""
        log_files = [f for (_, _, files) in os.walk(self.logs_dir_path) for f in files]

        for log_file in log_files:

            log_start_time, pid, process_type = parse_log_file_name(log_file)

            if self.process_type and process_type != self.process_type:
                continue

            self._log_map[log_file] = Log(
                log_start_time, pid, ProcessType(process_type)
            )

        self._log_map = OrderedDict(sorted(self._log_map.items()))

    def __iter__(self) -> Iterator[Log]:
        yield from self.log_list

    def get_from_file_name(self, file_name: str) -> Log:
        """Retrieve a `Log` instance based on the log file name.

        :param file_name: The name of the log file (with or without extension).
        :raises FileNotFoundError: If the log file name is not found.
        :return: The `Log` instance corresponding to the file name.
        """
        # auto strip the extension if present
        file_name, _ = os.path.splitext(file_name)
        try:
            return self._log_map[file_name]
        except KeyError:
            raise FileNotFoundError(
                f"Log handler for file name '{file_name}' not found in log map"
            )

    def get_from_pid(self, pid: str) -> Log:
        """Retrieve a `Log` instance based on the process ID.

        :param pid: The process ID to search for.
        :raises FileNotFoundError: If a log file corresponding to the process ID is not found.
        :return: The `Log` instance corresponding to the process ID.
        """
        for log in self.log_list:
            if log.pid == pid:
                return log
        raise FileNotFoundError(f"Log handler for PID '{pid}' could not be identified")



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Optional, TypeVar, Generic, Iterator, overload, cast

from ._pnames import PName

# value type
VT = TypeVar("VT")


class Parameter(Generic[VT]):
    """A simple container for a named value."""

    def __init__(self, name: PName, value: Optional[VT] = None) -> None:
        """Initialise a `Parameter` instance.

        :param name: The name of the parameter.
        :param value: The value of the parameter. Defaults to None.
        """
        self._name = name
        self._value: Optional[VT] = value

    @property
    def name(self) -> PName:
        """The parameter name."""
        return self._name

    @property
    def value(self) -> Optional[VT]:
        """The parameter value."""
        return self._value

    @value.setter
    def value(self, v: Optional[VT]) -> None:
        """Update the parameter value.

        :param v: The new value to set for the parameter.
        """
        self._value = v


class Parameters:
    """A managed collection of parameters."""

    def __init__(self) -> None:
        """Initialise a `Parameters` instance."""
        self._parameters: dict[PName, Parameter] = {}

    @property
    def name_list(self) -> list[PName]:
        """List the names of stored parameters."""
        return list(self._parameters.keys())

    def add_parameter(self, name: PName, value: Optional[VT]) -> None:
        """Add a `Parameter` instance to this `Parameters` instance with the input name and value.

        :param name: The name of the parameter.
        :param value: The value of the parameter.
        :raises KeyError: If a parameter already exists under the input name.
        """
        if name in self._parameters:
            raise KeyError(
                f"Cannot add a parameter with name '{name}', "
                f"since a parameter already exists with that name. "
            )
        self._parameters[name] = Parameter(name, value)

    def get_parameter(self, name: PName) -> Parameter:
        """Get the stored `Parameter` instance corresponding to the input name.

        :param name: The name of the parameter.
        :raises KeyError: If a parameter with the input name does not exist.
        :return: A `Parameter` instance with the input name, if it exists.
        """
        if name not in self._parameters:
            raise KeyError(
                f"Parameter with name '{name}' does not exist. "
                f"Expected one of {self.name_list}"
            )
        return self._parameters[name]

    def get_parameter_value(self, name: PName) -> Optional[VT]:
        """Get the value of the parameter with the corresponding name.

        :param name: The name of the parameter.
        :return: The value of the parameter with the input name.
        """
        return self.get_parameter(name).value

    def __iter__(self) -> Iterator[Parameter]:
        """Iterate over stored parameters."""
        yield from self._parameters.values()

    def to_dict(self) -> dict[str, Optional[Any]]:
        """Convert the `Parameters` instance to a serialisable dictionary.

        :return: A dictionary representation of the stored parameters.
        """
        return {p.name.value: p.value for p in self}


def _parse_string_parameter(string_parameter: str) -> list[str]:
    """Parse string of the form `a=b` into a list of the form `[a, b]`.

    :param string_parameter: A string representation of a capture config parameter.
    :raises ValueError: If the input parameter is not of the form `a=b`.
    :return: The parsed components of the input string parameter, using the `=` character as a separator.
    The return list will always contain two elements.
    """
    if not string_parameter or "=" not in string_parameter:
        raise ValueError(f"Invalid format: '{string_parameter}'. Expected 'KEY=VALUE'.")
    if string_parameter.startswith("=") or string_parameter.endswith("="):
        raise ValueError(f"Invalid format: '{string_parameter}'. Expected 'KEY=VALUE'.")
    # remove leading and trailing whitespace.
    string_parameter = string_parameter.strip()
    return string_parameter.split("=", 1)


def parse_string_parameters(string_parameters: list[str]) -> dict[str, str]:
    """Parses a list of strings of the form `a=b` into a dictionary mapping each `a` to each `b`.

    :param string_parameters: A list of strings, where each element is of the form `a=b`.
    :return: A dictionary mapping each `a` to each `b`, after parsing each element.
    """
    d = {}
    for string_parameter in string_parameters:
        k, v = _parse_string_parameter(string_parameter)
        d[k] = v
    return d


def make_parameters(d: dict[str, Any]) -> Parameters:
    """Create a `Parameters` instance from the given dictionary. Each key is interpreted as a valid `PName`.

    :param d: A dictionary where keys represent parameter names and values represent their corresponding values.
    :return: An instance of `Parameters` with each key-value pair in `d` added as a parameter.
    """
    parameters = Parameters()
    for k, v in d.items():
        parameters.add_parameter(PName(k), v)
    return parameters



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from functools import partial
from typing import Callable, Optional

from spectre_core.capture_configs import (
    CaptureTemplate,
    CaptureMode,
    Parameters,
    Bound,
    PName,
    OneOf,
    get_base_capture_template,
    get_base_ptemplate,
    validate_fixed_center_frequency,
    validate_swept_center_frequency,
    validate_sample_rate_with_master_clock_rate,
)

from ._receiver_names import ReceiverName
from ._gr import capture
from ._b200mini_gr import fixed_center_frequency, swept_center_frequency
from .._receiver import Receiver
from .._specs import SpecName
from .._register import register_receiver


def _make_pvalidator_fixed_center_frequency(
    receiver: Receiver,
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_fixed_center_frequency(parameters)
        validate_sample_rate_with_master_clock_rate(parameters)

    return pvalidator


def _make_pvalidator_swept_center_frequency(
    receiver: Receiver,
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_swept_center_frequency(
            parameters, receiver.get_spec(SpecName.API_RETUNING_LATENCY)
        )
        validate_sample_rate_with_master_clock_rate(parameters)

    return pvalidator


def _make_capture_template_fixed_center_frequency(
    receiver: Receiver,
) -> CaptureTemplate:

    capture_template = get_base_capture_template(CaptureMode.FIXED_CENTER_FREQUENCY)
    capture_template.add_ptemplate(get_base_ptemplate(PName.BANDWIDTH))
    capture_template.add_ptemplate(get_base_ptemplate(PName.GAIN))
    capture_template.add_ptemplate(get_base_ptemplate(PName.WIRE_FORMAT))
    capture_template.add_ptemplate(get_base_ptemplate(PName.MASTER_CLOCK_RATE))

    capture_template.set_defaults(
        (PName.BATCH_SIZE, 4.0),
        (PName.CENTER_FREQUENCY, 95800000),
        (PName.SAMPLE_RATE, 2000000),
        (PName.BANDWIDTH, 2000000),
        (PName.WINDOW_HOP, 512),
        (PName.WINDOW_SIZE, 1024),
        (PName.WINDOW_TYPE, "blackman"),
        (PName.GAIN, 35),
        (PName.WIRE_FORMAT, "sc16"),
        (PName.MASTER_CLOCK_RATE, 40e6),
    )

    capture_template.add_pconstraint(
        PName.CENTER_FREQUENCY,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.BANDWIDTH_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.BANDWIDTH_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.GAIN,
        [
            Bound(
                lower_bound=0,
                upper_bound=receiver.get_spec(SpecName.GAIN_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.WIRE_FORMAT, [OneOf(receiver.get_spec(SpecName.WIRE_FORMATS))]
    )
    capture_template.add_pconstraint(
        PName.MASTER_CLOCK_RATE,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.MASTER_CLOCK_RATE_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.MASTER_CLOCK_RATE_UPPER_BOUND),
            )
        ],
    )
    return capture_template


def _make_capture_template_swept_center_frequency(
    receiver: Receiver,
) -> CaptureTemplate:

    capture_template = get_base_capture_template(CaptureMode.SWEPT_CENTER_FREQUENCY)
    capture_template.add_ptemplate(get_base_ptemplate(PName.BANDWIDTH))
    capture_template.add_ptemplate(get_base_ptemplate(PName.GAIN))
    capture_template.add_ptemplate(get_base_ptemplate(PName.WIRE_FORMAT))
    capture_template.add_ptemplate(get_base_ptemplate(PName.MASTER_CLOCK_RATE))

    # TODO: Delegate defaults to receiver subclasses. Currently, these are sensible defaults for the b200mini
    capture_template.set_defaults(
        (PName.BATCH_SIZE, 4.0),
        (PName.MIN_FREQUENCY, 95000000),
        (PName.MAX_FREQUENCY, 105000000),
        (PName.SAMPLES_PER_STEP, 30000),
        (PName.FREQUENCY_STEP, 2000000),
        (PName.SAMPLE_RATE, 2000000),
        (PName.BANDWIDTH, 2000000),
        (PName.WINDOW_HOP, 512),
        (PName.WINDOW_SIZE, 1024),
        (PName.WINDOW_TYPE, "blackman"),
        (PName.GAIN, 35),
        (PName.WIRE_FORMAT, "sc16"),
        (PName.MASTER_CLOCK_RATE, 40e6),
    )

    capture_template.add_pconstraint(
        PName.MIN_FREQUENCY,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.MAX_FREQUENCY,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.BANDWIDTH_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.BANDWIDTH_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.GAIN,
        [
            Bound(
                lower_bound=0,
                upper_bound=receiver.get_spec(SpecName.GAIN_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.WIRE_FORMAT, [OneOf(receiver.get_spec(SpecName.WIRE_FORMATS))]
    )
    capture_template.add_pconstraint(
        PName.MASTER_CLOCK_RATE,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.MASTER_CLOCK_RATE_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.MASTER_CLOCK_RATE_UPPER_BOUND),
            )
        ],
    )
    return capture_template


@dataclass(frozen=True)
class _Mode:
    """An operating mode for the `B200mini` receiver."""

    FIXED_CENTER_FREQUENCY = "fixed_center_frequency"
    SWEPT_CENTER_FREQUENCY = "swept_center_frequency"


@register_receiver(ReceiverName.B200MINI)
class B200mini(Receiver):
    """Receiver implementation for the USRP B200mini (https://www.ettus.com/all-products/usrp-b200mini/)"""

    def __init__(self, name: ReceiverName, mode: Optional[str] = None) -> None:
        super().__init__(name, mode)

        self.add_spec(SpecName.SAMPLE_RATE_LOWER_BOUND, 200e3)
        self.add_spec(SpecName.SAMPLE_RATE_UPPER_BOUND, 56e6)
        self.add_spec(SpecName.FREQUENCY_LOWER_BOUND, 70e6)
        self.add_spec(SpecName.FREQUENCY_UPPER_BOUND, 6e9)
        self.add_spec(SpecName.BANDWIDTH_LOWER_BOUND, 200e3)
        self.add_spec(SpecName.BANDWIDTH_UPPER_BOUND, 56e6)
        self.add_spec(SpecName.GAIN_UPPER_BOUND, 76)
        self.add_spec(SpecName.WIRE_FORMATS, ["sc8", "sc12", "sc16"])
        self.add_spec(SpecName.MASTER_CLOCK_RATE_LOWER_BOUND, 5e6)
        self.add_spec(SpecName.MASTER_CLOCK_RATE_UPPER_BOUND, 61.44e6)
        self.add_spec(
            SpecName.API_RETUNING_LATENCY, 1e-5
        )  # TODO: This is a ballpark, pending empirical testing

        self.add_mode(
            _Mode.FIXED_CENTER_FREQUENCY,
            partial(capture, top_block_cls=fixed_center_frequency),
            _make_capture_template_fixed_center_frequency(self),
            _make_pvalidator_fixed_center_frequency(self),
        )

        self.add_mode(
            _Mode.SWEPT_CENTER_FREQUENCY,
            partial(capture, top_block_cls=swept_center_frequency),
            _make_capture_template_swept_center_frequency(self),
            _make_pvalidator_swept_center_frequency(self),
        )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional, overload, Literal

from spectre_core.exceptions import ReceiverNotFoundError
from ._register import receivers
from ._receiver import Receiver
from .plugins._receiver_names import ReceiverName
from .plugins._signal_generator import SignalGenerator
from .plugins._rsp1a import RSP1A
from .plugins._rspduo import RSPduo
from .plugins._b200mini import B200mini
from .plugins._custom import CustomReceiver


@overload
def get_receiver(
    receiver_name: Literal[ReceiverName.SIGNAL_GENERATOR], mode: Optional[str] = None
) -> SignalGenerator: ...


@overload
def get_receiver(
    receiver_name: Literal[ReceiverName.RSP1A], mode: Optional[str] = None
) -> RSP1A: ...


@overload
def get_receiver(
    receiver_name: Literal[ReceiverName.RSPDUO], mode: Optional[str] = None
) -> RSPduo: ...


@overload
def get_receiver(
    receiver_name: Literal[ReceiverName.B200MINI], mode: Optional[str] = None
) -> B200mini: ...


@overload
def get_receiver(
    receiver_name: Literal[ReceiverName.CUSTOM], mode: Optional[str] = None
) -> CustomReceiver: ...


@overload
def get_receiver(
    receiver_name: ReceiverName, mode: Optional[str] = None
) -> Receiver: ...


def get_receiver(receiver_name: ReceiverName, mode: Optional[str] = None) -> Receiver:
    """Get a registered receiver.

    :param receiver_name: The name of the receiver.
    :param mode: The initial operating mode for the receiver, defaults to None
    :raises ReceiverNotFoundError: If the receiver name is not registered.
    :return: An instance of the receiver class registered under `receiver_name`.
    """
    receiver_cls = receivers.get(receiver_name)
    if receiver_cls is None:
        valid_receivers = list(receivers.keys())
        raise ReceiverNotFoundError(
            f"No class found for the receiver: {receiver_name}. "
            f"Please specify one of the following receivers {valid_receivers}"
        )
    return receiver_cls(receiver_name, mode=mode)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


"""General `spectre` package configurations."""

import os

from ._paths import (
    get_spectre_data_dir_path,
    get_batches_dir_path,
    get_configs_dir_path,
    get_logs_dir_path,
    trim_spectre_data_dir_path,
)
from ._time_formats import TimeFormat

os.makedirs(get_batches_dir_path(), exist_ok=True)
os.makedirs(get_logs_dir_path(), exist_ok=True)
os.makedirs(get_configs_dir_path(), exist_ok=True)

__all__ = [
    "get_spectre_data_dir_path",
    "get_batches_dir_path",
    "get_configs_dir_path",
    "get_logs_dir_path",
    "TimeFormat",
    "trim_spectre_data_dir_path",
]



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import logging
from typing import Tuple
from datetime import datetime

from spectre_core.config import TimeFormat
from ._logs import Log
from ._process_types import ProcessType


def configure_root_logger(process_type: ProcessType, level: int = logging.INFO) -> str:
    """Configures the root logger to write logs to a file named based on
    the process type, process ID, and the current system time.

    :param process_type: Indicates the type of process, as defined by `ProcessType`.
    :param level: The logging level, as defined in Python's `logging` module. Defaults to `logging.INFO`.
    :return: The file path of the created log file.
    """
    # create a `spectre` log handler instance, to represent the log file.
    # get the star time of the log
    system_datetime = datetime.now()
    start_time = system_datetime.strftime(TimeFormat.DATETIME)

    # extract the process identifier, and cast as a string
    pid = str(os.getpid())

    # create a file handler representing the log file
    log = Log(start_time, pid, process_type)
    log.make_parent_dir_path()

    # get the root logger and set its level.
    logger = logging.getLogger()
    logger.setLevel(level)

    # remove existing handlers
    for handler in logger.handlers:
        logger.removeHandler(handler)

    # Set up a file handler and add it to the root logger
    file_handler = logging.FileHandler(log.file_path)
    file_handler.setLevel(level)
    formatter = logging.Formatter(
        "[%(asctime)s] [%(levelname)8s] --- %(message)s (%(name)s:%(lineno)s)"
    )
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    return log.file_path


def get_root_logger_state() -> Tuple[bool, int]:
    """Get the state of the root logger.

    :return: Whether the root logger has any handlers, and the level of the root logger.
    """
    root_logger = logging.getLogger()
    if root_logger.handlers:
        return True, root_logger.level
    return False, logging.NOTSET



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Type, cast

from spectre_core.post_processing._base import BaseEventHandler
from spectre_core.capture_configs import CaptureConfig, PName
from spectre_core.exceptions import EventHandlerNotFoundError
from .plugins._event_handler_keys import EventHandlerKey
from ._register import event_handler_map


def _get_event_handler_cls_from_key(
    event_handler_key: EventHandlerKey,
) -> Type[BaseEventHandler]:
    """Get a registered `BaseEventHandler` subclass.

    :param event_handler_key: The key used to register the `BaseEventHandler` subclass.
    :raises EventHandlerNotFoundError: If an invalid `event_handler_key` is provided.
    :return: The `BaseEventHandler` subclass registered under `event_handler_key`.
    """
    event_handler_cls = event_handler_map.get(event_handler_key)
    if event_handler_cls is None:
        event_handler_keys = list(event_handler_map.keys())
        raise EventHandlerNotFoundError(
            (
                f"No event handler found for the capture mode '{event_handler_key}'. "
                f"Please specify one of the following capture modes: {event_handler_keys}"
            )
        )
    return event_handler_cls


def get_event_handler_cls_from_tag(tag: str) -> Type[BaseEventHandler]:
    """
    Retrieve the event handler class, using the event handler key stored in a capture config.

    :param tag: The capture config tag.
    :return: The event handler class specified in the capture config.
    """
    capture_config = CaptureConfig(tag)
    event_handler_key = cast(
        str, capture_config.get_parameter_value(PName.EVENT_HANDLER_KEY)
    )
    return _get_event_handler_cls_from_key(EventHandlerKey(event_handler_key))


def get_event_handler(tag: str) -> BaseEventHandler:
    """Create an event handler class instance, using the event handler key stored in a capture config.

    :param tag: The capture config tag.
    :return: An instance of the event handler class.
    """
    event_handler_cls = get_event_handler_cls_from_tag(tag)
    return event_handler_cls(tag)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger

_LOGGER = getLogger(__name__)

from typing import Optional, cast
from abc import ABC, abstractmethod
from scipy.signal import ShortTimeFFT, get_window

from watchdog.events import FileSystemEventHandler, FileSystemEvent

from spectre_core.capture_configs import CaptureConfig, PName
from spectre_core.spectrograms import Spectrogram, join_spectrograms


def make_sft_instance(capture_config: CaptureConfig) -> ShortTimeFFT:
    """Extract window parameters from the input capture config and create an instance
    of `ShortTimeFFT` from `scipy.signal`.

    :param capture_config: The capture config storing the parameters.
    :return: An instance of `ShortTimeFFT` consistent with the window parameters
    in the capture config.
    """
    window_type = cast(str, capture_config.get_parameter_value(PName.WINDOW_TYPE))
    sample_rate = cast(int, capture_config.get_parameter_value(PName.SAMPLE_RATE))
    window_hop = cast(int, capture_config.get_parameter_value(PName.WINDOW_HOP))
    window_size = cast(int, capture_config.get_parameter_value(PName.WINDOW_SIZE))
    window = get_window(window_type, window_size)
    return ShortTimeFFT(window, window_hop, sample_rate, fft_mode="centered")


class BaseEventHandler(ABC, FileSystemEventHandler):
    """An abstract base class for event-driven file post-processing."""

    def __init__(self, tag: str) -> None:
        """Initialise a `BaseEventHandler` instance.

        :param tag: The tag of the capture config used to capture the data.
        """
        self._tag = tag

        # load the capture config corresponding to the input tag
        self._capture_config = CaptureConfig(tag)

        # store the next file to be processed (specifically, the absolute file path of the file)
        self._queued_file: Optional[str] = None

        # optionally store batched spectrograms as they are created into a cache
        # this can be flushed periodically to file as required.
        self._cached_spectrogram: Optional[Spectrogram] = None

    @abstractmethod
    def process(self, absolute_file_path: str) -> None:
        """
        Process a batch file at the given file path.

        :param absolute_file_path: The absolute path to the batch file to be processed.
        """

    def on_created(self, event: FileSystemEvent) -> None:
        """Process a newly created batch file, only once the next batch is created.

        Since we assume that the batches are non-overlapping in time, this guarantees
        we avoid post processing a file while it is being written to. Files are processed
        sequentially, in the order they are created.

        :param event: The file system event containing the file details.
        """
        # The `src_path`` attribute holds the absolute path of the freshly closed file
        absolute_file_path = event.src_path

        # Only process a file if:
        #
        # - It's extension matches the `watch_extension` as defined in the capture config.
        # - It's tag matches the current sessions tag.
        #
        # This is important for two reasons.
        #
        # In the case of one session, the capture worker may write to two batch files simultaneously
        # (e.g., raw data file + seperate metadata file). We want to process them together - but this method will get called
        # seperately for both file creation events. So, we filter by extension to account for this.
        #
        # Additionally in the case of multiple sessions, the capture workers will create batch files in the same directory concurrently.
        # This method is triggered for all file creation events, so we ensure the batch file tag matches the session tag and early return
        # otherwise. This way, each post processor worker picks up the right files to process.
        watch_extension = cast(
            str, self._capture_config.get_parameter_value(PName.WATCH_EXTENSION)
        )
        if not absolute_file_path.endswith(f"{self._tag}.{watch_extension}"):
            return

        _LOGGER.info(f"Noticed {absolute_file_path}")
        # If there exists a queued file, try and process it
        if self._queued_file is not None:
            try:
                self.process(self._queued_file)
            except Exception:
                _LOGGER.error(
                    f"An error has occured while processing {self._queued_file}",
                    exc_info=True,
                )
                # Flush any internally stored spectrogram on error to avoid lost data
                self._flush_cache()
                # re-raise the exception to the main thread
                raise

        # Queue the current file for processing next
        _LOGGER.info(f"Queueing {absolute_file_path} for post processing")
        self._queued_file = absolute_file_path

    def _cache_spectrogram(self, spectrogram: Spectrogram) -> None:
        """Cache the input spectrogram by storing it in the `_cached_spectrogram` attribute.

        If the time range of the cached spectrogram exceeds that as specified in the capture config
        `PName.TIME_RANGE` parameter, the spectrogram in the cache is flushed to file. If `PName.TIME_RANGE`
        is nulled, the cache is flushed immediately.

        :param spectrogram: The spectrogram to store in the cache.
        """
        _LOGGER.info("Joining spectrogram")

        if self._cached_spectrogram is None:
            self._cached_spectrogram = spectrogram
        else:
            self._cached_spectrogram = join_spectrograms(
                [self._cached_spectrogram, spectrogram]
            )

        time_range = self._capture_config.get_parameter_value(PName.TIME_RANGE) or 0.0
        if self._cached_spectrogram.time_range >= cast(float, time_range):
            self._flush_cache()

    def _flush_cache(self) -> None:
        """Flush the cached spectrogram to file."""
        if self._cached_spectrogram:
            _LOGGER.info(
                f"Flushing spectrogram to file with start time "
                f"'{self._cached_spectrogram.format_start_time()}'"
            )
            self._cached_spectrogram.save()
            _LOGGER.info("Flush successful, resetting spectrogram cache")
            self._cached_spectrogram = None  # reset the cache



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Optional, Literal
from enum import Enum
from datetime import datetime

import numpy.typing as npt
import numpy as np
import matplotlib.dates as mdates
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from spectre_core.spectrograms import Spectrogram, TimeType
from spectre_core.config import TimeFormat
from ._format import PanelFormat
from ._panel_names import PanelName


class XAxisType(Enum):
    """The xaxis type for a panel.

    Axes are shared in a stack between panels with common `XAxisType`.

    :ivar TIME: The xaxis has units of time.
    :ivar FREQUENCY: The xaxis has units of frequency.
    """

    TIME = "time"
    FREQUENCY = "frequency"


class BasePanel(ABC):
    """Abstract base class for a panel used to visualise spectrogram data.

    `BasePanel` instances are designed to be part of a `PanelStack`, where multiple
    panels contribute to a composite plot. Subclasses must implement methods to define
    how the panel is drawn and annotated, and specify its xaxis type.
    """

    def __init__(
        self,
        name: PanelName,
        spectrogram: Spectrogram,
        time_type: TimeType = TimeType.RELATIVE,
    ) -> None:
        """Initialize an instance of `BasePanel`.

        :param name: The name of the panel.
        :param spectrogram: The spectrogram being visualised.
        :param time_type: Indicates whether the times of each spectrum are relative to the first
        spectrum in the spectrogram, or datetimes.
        """
        self._name = name
        self._spectrogram = spectrogram
        self._time_type = time_type

        # These attributes should be set by instances of `PanelStack`.
        self._ax: Optional[Axes] = None
        self._fig: Optional[Figure] = None
        self._panel_format: Optional[PanelFormat] = None
        self._identifier: Optional[str] = None

    @abstractmethod
    def draw(self) -> None:
        """Modify the `ax` attribute to draw the panel contents."""

    @abstractmethod
    def annotate_xaxis(self) -> None:
        """Modify the `ax` attribute to annotate the xaxis of the panel."""

    @abstractmethod
    def annotate_yaxis(self) -> None:
        """Modify the `ax` attribute to annotate the yaxis of the panel."""

    @property
    @abstractmethod
    def xaxis_type(self) -> XAxisType:
        """Specify the xaxis type for the panel."""

    @property
    def spectrogram(self) -> Spectrogram:
        """The spectrogram being visualised on this panel."""
        return self._spectrogram

    @property
    def tag(self) -> str:
        """The tag of the spectrogram being visualised."""
        return self._spectrogram.tag

    @property
    def name(self) -> PanelName:
        """The name of the panel."""
        return self._name

    def get_time_type(self) -> TimeType:
        """The time type of the spectrogram.

        :raises ValueError: If the `time_type` has not been set.
        """
        return self._time_type

    def set_time_type(self, value: TimeType) -> None:
        """Set the `TimeType` for the spectrogram.

        This controls how time is represented and annotated on the panel.

        :param value: The `TimeType` to assign to the spectrogram.
        """
        self._time_type = value

    def get_panel_format(self) -> PanelFormat:
        """Retrieve the panel format, which controls the style of the panel.

        :raises ValueError: If the `panel_format` has not been set.
        """
        if self._panel_format is None:
            raise ValueError(f"`panel_format` must be set for the panel `{self.name}`")
        return self._panel_format

    def set_panel_format(self, value: PanelFormat) -> None:
        """Set the panel format to control the style of the panel.

        :param value: The `PanelFormat` to assign to the panel.
        """
        self._panel_format = value

    def _get_ax(self) -> Axes:
        """Return the `Axes` object bound to this panel.

        This method is protected to restrict direct access to `matplotlib` functionality,
        promoting encapsulation, promoting encapsulation.

        :raises ValueError: If the `Axes` object has not been set.
        """
        if self._ax is None:
            raise ValueError(f"`ax` must be set for the panel `{self.name}`")
        return self._ax

    def set_ax(self, value: Axes) -> None:
        """Assign a Matplotlib `Axes` object to this panel.

        This `Axes` will be used for drawing and annotations.

        :param value: The Matplotlib `Axes` to assign to the panel.
        """
        self._ax = value

    def _get_fig(self) -> Figure:
        """Return the `Figure` object bound to this panel.

        This method is protected to restrict direct access to `matplotlib` functionality,
        promoting encapsulation, promoting encapsulation.

        :raises ValueError: If the `Figure` object has not been set.
        """
        if self._fig is None:
            raise ValueError(f"`fig` must be set for the panel `{self.name}`")
        return self._fig

    def set_fig(self, value: Figure) -> None:
        """
        Assign a Matplotlib `Figure` object to this panel.

        This `Figure` is shared across all panels in the `PanelStack`.

        :param value: The Matplotlib `Figure` to assign to the panel.
        """
        self._fig = value

    def get_identifier(self) -> Optional[str]:
        """Optional identifier for the panel.

        This identifier can be used to distinguish panels or aid in superimposing
        panels in a stack.
        """
        return self._identifier

    def set_identifier(self, value: str) -> None:
        """Set the optional identifier for the panel.

        This can be used to distinguish panels or aid in superimposing panels.
        """
        self._identifier = value

    def hide_xaxis_labels(self) -> None:
        """Hide the labels for xaxis ticks in the panel."""
        self._get_ax().tick_params(axis="x", labelbottom=False)

    def hide_yaxis_labels(self) -> None:
        """Hide the labels for yaxis ticks in the panel."""
        self._get_ax().tick_params(axis="y", labelleft=False)

    def sharex(self, axes: Axes):
        """Share the xaxis with another axes."""
        self._get_ax().sharex(axes)

    def get_xaxis_labels(self) -> list[str]:
        """Get the text string of all xaxis tick labels."""
        tick_labels = self._get_ax().get_xaxis().get_ticklabels(which="both")
        return [tick_label.get_text() for tick_label in tick_labels]

    def get_yaxis_labels(self) -> list[str]:
        """Get the text string of all yaxis tick labels."""
        tick_labels = self._get_ax().get_yaxis().get_ticklabels(which="both")
        return [tick_label.get_text() for tick_label in tick_labels]

    def get_xlabel(self) -> str:
        """Get the xlabel text string."""
        return self._get_ax().get_xlabel()

    def get_ylabel(self) -> str:
        """Get the ylabel text string."""
        return self._get_ax().get_ylabel()

    def share_axes(self, panel: "BasePanel") -> None:
        # TODO: More elegantly share axes, rather than access protected method from
        # another instance. Probably, this should just be entirely managed by the `PanelStack`
        self.set_ax(panel._get_ax())
        self.set_fig(panel._get_fig())


class BaseTimeSeriesPanel(BasePanel):
    """
    Abstract subclass of `BasePanel` designed for visualising time series data.

    Subclasses must implement any remaining abstract methods from `BasePanel`.
    """

    @property
    def xaxis_type(self) -> Literal[XAxisType.TIME]:
        return XAxisType.TIME

    @property
    def times(self) -> npt.NDArray[np.float32 | np.datetime64]:
        """The times assigned to each spectrum according to the `TimeType`."""
        return (
            self.spectrogram.times
            if self.get_time_type() == TimeType.RELATIVE
            else self.spectrogram.datetimes
        )

    def annotate_xaxis(self) -> None:
        """Annotate the xaxis according to the specified `TimeType`."""
        ax = self._get_ax()
        if self.get_time_type() == TimeType.RELATIVE:
            ax.set_xlabel("Time [s]")
        else:
            # TODO: Adapt for time ranges greater than one day
            start_date = datetime.strftime(
                self.spectrogram.start_datetime.astype(datetime), TimeFormat.DATE
            )
            ax.set_xlabel(f"Time [UTC] (Start Date: {start_date})")
            ax.xaxis.set_major_formatter(mdates.DateFormatter(TimeFormat.TIME))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from typing import Optional, TypeVar, Type, Generic, Iterator
from collections import OrderedDict
from datetime import datetime

from spectre_core.config import TimeFormat
from spectre_core.spectrograms import Spectrogram, time_chop, join_spectrograms
from spectre_core.config import get_batches_dir_path
from spectre_core.exceptions import BatchNotFoundError
from ._base import BaseBatch, parse_batch_file_name

T = TypeVar("T", bound=BaseBatch)


class Batches(Generic[T]):
    """Managed collection of `Batch` instances for a given tag. Provides a simple
    interface for read operations on batched data files."""

    def __init__(
        self,
        tag: str,
        batch_cls: Type[T],
        year: Optional[int] = None,
        month: Optional[int] = None,
        day: Optional[int] = None,
    ) -> None:
        """Initialise a `Batches` instance.

        :param tag: The batch name tag.
        :param batch_cls: The `Batch` class used to read data files tagged by `tag`.
        :param year: Filter batch files under a numeric year. Defaults to None.
        :param month: Filter batch files under a numeric month. Defaults to None.
        :param day: Filter batch files under a numeric day. Defaults to None.
        """
        self._tag = tag
        self._batch_cls = batch_cls
        self._batch_map: dict[str, T] = OrderedDict()
        self.set_date(year, month, day)

    @property
    def tag(self) -> str:
        """The batch name tag."""
        return self._tag

    @property
    def batch_cls(self) -> Type[T]:
        """The `Batch` class used to read the batched files."""
        return self._batch_cls

    @property
    def year(self) -> Optional[int]:
        """The numeric year, to filter batch files."""
        return self._year

    @property
    def month(self) -> Optional[int]:
        """The numeric month of the year, to filter batch files."""
        return self._month

    @property
    def day(self) -> Optional[int]:
        """The numeric day of the year, to filter batch files."""
        return self._day

    @property
    def batches_dir_path(self) -> str:
        """The shared ancestral path for all the batches. `Batches` recursively searches
        this directory to find all batches whose batch name contains `tag`."""
        return get_batches_dir_path(self.year, self.month, self.day)

    @property
    def batch_list(self) -> list[T]:
        """A list of all batches found within `batches_dir_path`."""
        return list(self._batch_map.values())

    @property
    def start_times(self) -> list[str]:
        """The start times of each batch found within `batches_dir_path`."""
        return list(self._batch_map.keys())

    @property
    def num_batches(self) -> int:
        """The total number of batches found within `batches_dir_path`."""
        return len(self.batch_list)

    def set_date(
        self, year: Optional[int], month: Optional[int], day: Optional[int]
    ) -> None:
        """Reset `batches_dir_path` according to the numeric date, and refresh the list
        of available batches.

        :param year: Filter by the numeric year.
        :param month: Filter by the numeric month of the year.
        :param day: Filter by the numeric day of the month.
        """
        self._year = year
        self._month = month
        self._day = day
        self.update()

    def update(self) -> None:
        """Perform a fresh search all files in `batches_dir_path` for batches
        with `tag` in the batch name."""
        # reset cache
        self._batch_map = OrderedDict()

        # get a list of all batch file names in the batches directory path
        batch_file_names = [
            f for (_, _, files) in os.walk(self.batches_dir_path) for f in files
        ]
        for batch_file_name in batch_file_names:
            start_time, tag, _ = parse_batch_file_name(batch_file_name)
            if tag == self._tag:
                self._batch_map[start_time] = self.batch_cls(start_time, tag)

        self._batch_map = OrderedDict(sorted(self._batch_map.items()))

    def __iter__(self) -> Iterator[T]:
        """Iterate over the stored batch instances."""
        yield from self.batch_list

    def __len__(self):
        return self.num_batches

    def _get_from_start_time(self, start_time: str) -> T:
        """Find and return the `Batch` instance based on the string formatted start time."""
        try:
            return self._batch_map[start_time]
        except KeyError:
            raise BatchNotFoundError(
                f"Batch with start time {start_time} could not be found within {self.batches_dir_path}"
            )

    def _get_from_index(self, index: int) -> T:
        """Find and return the `Batch` instance based on its numeric index.

        Batches are ordered sequentially in time, so index `0` corresponds to the oldest
        `Batch` with respect to the start time.
        """
        if self.num_batches == 0:
            raise BatchNotFoundError("No batches are available")
        elif index > self.num_batches:
            raise IndexError(
                f"Index '{index}' is greater than the number of batches '{self.num_batches}'"
            )
        return self.batch_list[index]

    def __getitem__(self, subscript: str | int) -> T:
        """Get a `Batch` instanced based on either the start time or chronological index.

        :param subscript: If the subscript is a string, interpreted as a formatted start time.
        If the subscript is an integer, it is interpreted as a chronological index.
        :return: The corresponding `BaseBatch` subclass.
        """
        if isinstance(subscript, str):
            return self._get_from_start_time(subscript)
        elif isinstance(subscript, int):
            return self._get_from_index(subscript)

    def get_spectrogram(
        self, start_datetime: datetime, end_datetime: datetime
    ) -> Spectrogram:
        """
        Retrieve a spectrogram spanning the specified time range.

        :param start_datetime: The start time of the range (inclusive).
        :param end_datetime: The end time of the range (inclusive).
        :raises FileNotFoundError: If no spectrogram data is available within the specified time range.
        :return: A spectrogram created by stitching together data from all matching batches.
        """

        spectrograms = []
        for batch in self:
            # skip batches without spectrogram data
            if not batch.spectrogram_file.exists:
                continue

            spectrogram = batch.read_spectrogram()
            lower_bound = spectrogram.datetimes[0]
            upper_bound = spectrogram.datetimes[-1]

            # Check if the batch overlaps with the input time range
            if start_datetime <= upper_bound and lower_bound <= end_datetime:
                spectrograms.append(
                    time_chop(spectrogram, start_datetime, end_datetime)
                )

        if spectrograms:
            return join_spectrograms(spectrograms)
        else:
            raise FileNotFoundError(
                f"No spectrogram data found for the time range "
                f"{start_datetime} to {end_datetime}."
            )



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from gnuradio import gr
from gnuradio import blocks
from gnuradio import analog
from gnuradio import spectre

from spectre_core.capture_configs import Parameters, PName
from spectre_core.config import get_batches_dir_path
from ._gr import spectre_top_block


class cosine_wave(spectre_top_block):
    def flowgraph(self, tag: str, parameters: Parameters) -> None:

        # Unpack the capture config parameters
        samp_rate = parameters.get_parameter_value(PName.SAMPLE_RATE)
        batch_size = parameters.get_parameter_value(PName.BATCH_SIZE)
        frequency = parameters.get_parameter_value(PName.FREQUENCY)
        amplitude = parameters.get_parameter_value(PName.AMPLITUDE)

        # Blocks
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            get_batches_dir_path(), tag, batch_size, samp_rate
        )
        self.blocks_throttle_0 = blocks.throttle(gr.sizeof_float * 1, samp_rate, True)
        self.blocks_throttle_1 = blocks.throttle(gr.sizeof_float * 1, samp_rate, True)
        self.blocks_null_source = blocks.null_source(gr.sizeof_float * 1)
        self.blocks_float_to_complex = blocks.float_to_complex(1)
        self.analog_sig_source = analog.sig_source_f(
            samp_rate, analog.GR_COS_WAVE, frequency, amplitude, 0, 0
        )

        # Connections
        self.connect((self.analog_sig_source, 0), (self.blocks_throttle_0, 0))
        self.connect((self.blocks_null_source, 0), (self.blocks_throttle_1, 0))
        self.connect((self.blocks_throttle_0, 0), (self.blocks_float_to_complex, 0))
        self.connect((self.blocks_throttle_1, 0), (self.blocks_float_to_complex, 1))
        self.connect(
            (self.blocks_float_to_complex, 0), (self.spectre_batched_file_sink, 0)
        )


class constant_staircase(spectre_top_block):
    def flowgraph(self, tag: str, parameters: Parameters) -> None:
        # Unpack the capture config parameters
        step_increment = parameters.get_parameter_value(PName.STEP_INCREMENT)
        samp_rate = parameters.get_parameter_value(PName.SAMPLE_RATE)
        min_samples_per_step = parameters.get_parameter_value(
            PName.MIN_SAMPLES_PER_STEP
        )
        max_samples_per_step = parameters.get_parameter_value(
            PName.MAX_SAMPLES_PER_STEP
        )
        frequency_step = parameters.get_parameter_value(PName.FREQUENCY_STEP)
        batch_size = parameters.get_parameter_value(PName.BATCH_SIZE)

        # Blocks
        self.spectre_constant_staircase = spectre.tagged_staircase(
            min_samples_per_step,
            max_samples_per_step,
            frequency_step,
            step_increment,
            samp_rate,
        )
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            get_batches_dir_path(), tag, batch_size, samp_rate, True, "rx_freq", 0
        )  # zero means the center frequency is unset
        self.blocks_throttle = blocks.throttle(
            gr.sizeof_gr_complex * 1, samp_rate, True
        )

        # Connections
        self.connect((self.spectre_constant_staircase, 0), (self.blocks_throttle, 0))
        self.connect((self.blocks_throttle, 0), (self.spectre_batched_file_sink, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from math import floor
from typing import Optional, cast

from scipy.signal import get_window

from ._parameters import Parameters
from ._pnames import PName


def validate_window(parameters: Parameters) -> None:
    """Ensure that the capture config describes a valid window.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the window interval is greater than the batch size.
    :raises ValueError: If the specified window type cannot be fetched using
    the SciPy `get_window` function.
    """
    window_size = cast(int, parameters.get_parameter_value(PName.WINDOW_SIZE))
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
    batch_size = cast(int, parameters.get_parameter_value(PName.BATCH_SIZE))
    window_type = cast(str, parameters.get_parameter_value(PName.WINDOW_TYPE))

    window_interval = window_size * (1 / sample_rate)
    if window_interval > batch_size:
        raise ValueError(
            (
                f"The windowing interval must be strictly less than the batch size. "
                f"Computed the windowing interval to be {window_interval} [s], "
                f"but the batch size is {batch_size} [s]"
            )
        )

    try:
        _ = get_window(window_type, window_size)
    except Exception as e:
        raise ValueError(
            (f"An error has occurred while validating the window. " f"Got {str(e)}")
        )


def validate_nyquist_criterion(parameters: Parameters) -> None:
    """Ensure that the Nyquist criterion is satisfied.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the sample rate is less than the bandwidth.
    """
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
    bandwidth = cast(float, parameters.get_parameter_value(PName.BANDWIDTH))

    if sample_rate < bandwidth:
        raise ValueError(
            (
                f"Nyquist criterion has not been satisfied. "
                f"Sample rate must be greater than or equal to the bandwidth. "
                f"Got sample rate {sample_rate} [Hz], and bandwidth {bandwidth} [Hz]"
            )
        )


def _compute_num_steps_per_sweep(
    min_freq: float, max_freq: float, freq_step: float
) -> int:
    """Compute the number of steps in one frequency sweep.

    The center frequency starts at `min_freq` and increments in steps of `freq_step`
    until the next step would exceed `max_freq`.

    :param min_freq: The minimum frequency of the sweep.
    :param max_freq: The maximum frequency of the sweep.
    :param freq_step: The frequency step size.
    :return: The number of steps in one frequency sweep.
    """
    return floor((max_freq - min_freq) / freq_step)


def validate_num_steps_per_sweep(parameters: Parameters) -> None:
    """Ensure that there are at least two steps in frequency per sweep.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the number of steps per sweep is less than or equal to one.
    """
    min_freq = cast(float, parameters.get_parameter_value(PName.MIN_FREQUENCY))
    max_freq = cast(float, parameters.get_parameter_value(PName.MAX_FREQUENCY))
    freq_step = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))

    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, max_freq, freq_step)
    if num_steps_per_sweep <= 1:
        raise ValueError(
            (
                f"We need strictly greater than one step per sweep. "
                f"Computed {num_steps_per_sweep} step per sweep"
            )
        )


def validate_sweep_interval(parameters: Parameters) -> None:
    """Ensure that the sweep interval is greater than the batch size.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the sweep interval is greater than the batch size.
    """
    min_freq = cast(float, parameters.get_parameter_value(PName.MIN_FREQUENCY))
    max_freq = cast(float, parameters.get_parameter_value(PName.MAX_FREQUENCY))
    freq_step = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))
    samples_per_step = cast(int, parameters.get_parameter_value(PName.SAMPLES_PER_STEP))
    batch_size = cast(int, parameters.get_parameter_value(PName.BATCH_SIZE))
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, max_freq, freq_step)
    num_samples_per_sweep = num_steps_per_sweep * samples_per_step
    sweep_interval = num_samples_per_sweep * 1 / sample_rate
    if sweep_interval > batch_size:
        raise ValueError(
            (
                f"Sweep interval must be less than the batch size. "
                f"The computed sweep interval is {sweep_interval} [s], "
                f"but the given batch size is {batch_size} [s]"
            )
        )


def validate_num_samples_per_step(parameters: Parameters) -> None:
    """Ensure that the number of samples per step is greater than the window size.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the window size is greater than the number of samples per step.
    """
    window_size = cast(int, parameters.get_parameter_value(PName.WINDOW_SIZE))
    samples_per_step = cast(int, parameters.get_parameter_value(PName.SAMPLES_PER_STEP))

    if window_size >= samples_per_step:
        raise ValueError(
            (
                f"Window size must be strictly less than the number of samples per step. "
                f"Got window size {window_size} [samples], which is more than or equal "
                f"to the number of samples per step {samples_per_step}"
            )
        )


def validate_non_overlapping_steps(parameters: Parameters) -> None:
    """Ensure that the stepped spectrograms are non-overlapping in the frequency domain.

    :param parameters: The parameters to be validated.
    :raises NotImplementedError: If the spectrograms overlap in the frequency domain.
    """

    freq_step = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    if freq_step < sample_rate:
        raise NotImplementedError(
            f"SPECTRE does not yet support spectral steps overlapping in frequency. "
            f"Got frequency step {freq_step * 1e-6} [MHz] which is less than the sample "
            f"rate {sample_rate * 1e-6} [MHz]"
        )


def validate_step_interval(parameters: Parameters, api_retuning_latency: float) -> None:
    """Ensure that the time elapsed collecting samples at a fixed frequency is greater
    than the empirically derived API retuning latency.

    :param parameters: The parameters to be validated.
    :param api_retuning_latency: The empirically derived API retuning latency (in seconds).
    :raises ValueError: If the time elapsed for a step is less than the API retuning latency.
    """
    samples_per_step = cast(int, parameters.get_parameter_value(PName.SAMPLES_PER_STEP))
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    step_interval = samples_per_step * 1 / sample_rate  # [s]
    if step_interval < api_retuning_latency:
        raise ValueError(
            f"The computed step interval of {step_interval} [s] is of the order of empirically "
            f"derived api latency {api_retuning_latency} [s]; you may experience undefined behaviour!"
        )


def validate_sample_rate_with_master_clock_rate(
    parameters: Parameters,
) -> None:
    """Ensure that the master clock rate is an integer multiple of the sample rate.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the master clock rate is not an integer multiple of the sample rate
    """
    master_clock_rate = cast(
        int, parameters.get_parameter_value(PName.MASTER_CLOCK_RATE)
    )
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    if master_clock_rate % sample_rate != 0:
        raise ValueError(
            f"The master clock rate of {master_clock_rate} [Hz] is not an integer "
            f"multiple of the sample rate {sample_rate} [Hz]."
        )


def validate_fixed_center_frequency(parameters: Parameters) -> None:
    """Apply validators for capture config parameters describing fixed center frequency capture.

    :param parameters: The parameters to be validated.
    """
    validate_nyquist_criterion(parameters)
    validate_window(parameters)


def validate_swept_center_frequency(
    parameters: Parameters,
    api_retuning_latency: Optional[float] = None,
) -> None:
    """Apply validators for capture config parameters describing swept center frequency capture.

    :param parameters: The parameters to be validated.
    :param api_retuning_latency: The empirically derived API retuning latency. Defaults to None.
    """
    validate_nyquist_criterion(parameters)
    validate_window(parameters)
    validate_non_overlapping_steps(parameters)
    validate_num_steps_per_sweep(parameters)
    validate_num_samples_per_step(parameters)
    validate_sweep_interval(parameters)

    if api_retuning_latency is not None:
        validate_step_interval(parameters, api_retuning_latency)



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

# TODO:
# - Make sure docstrings are filled in.
# - Make the tests more clear... Some seem opaque, and rely on particular properties of the fixtures.
# - Write a test for saving the batch file to a temp directory and making sure the filename is right etc.

import pytest
import tempfile
import numpy as np
import os
from datetime import datetime, timedelta
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from spectre_core.plotting import (
    BasePanel,
    XAxisType,
    PanelName,
    SpectrogramPanel,
    BaseTimeSeriesPanel,
    FrequencyCutsPanel,
    TimeCutsPanel,
    IntegralOverFrequencyPanel,
    PanelStack,
)
from spectre_core.config import TimeFormat
from spectre_core.spectrograms import Spectrogram, SpectrumUnit, TimeType


@pytest.fixture(autouse=True)
def setup_random_seed():
    """Set fixed random seed for all tests in this module."""
    # Arbitrarily chosen random seed.
    np.random.seed(42)


@pytest.fixture
def figure() -> Figure:
    return Figure()


@pytest.fixture
def axes(figure: Figure) -> Axes:
    return Axes(figure, [0.1, 0.1, 0.8, 0.8])


# An arbitrary datetime to assign to the first spectrum in the spectrogram fixture.
_ARBITRARY_DATETIME = datetime(2025, 2, 13, 6, 0, 0)


@pytest.fixture
def spectrogram() -> Spectrogram:
    """Create a spectrogram, using random values for the spectral components.

    This is used to instantiate some panels.
    """
    num_spectrums = 20
    num_spectral_components = 64

    # Populate the dynamic spectra with some random values.
    dynamic_spectra = np.random.uniform(
        -1, 1, (num_spectral_components, num_spectrums)
    ).astype(np.float32)
    # Arbitrarily, consider `num_spectrums` spectrums over an interval of 10 [s]
    times = np.linspace(0, 10, num_spectrums).astype(np.float32)
    # Arbitrarily, consider `num_spectral_components` spectral components, over the FM band
    frequencies = np.linspace(90e6, 110e6, num_spectral_components).astype(np.float32)
    # Arbitrarily, assign some date to be the datetime corresponding to the first spectrum.
    start_datetime = _ARBITRARY_DATETIME
    # Arbitrarily, set the unit (the plotting module does not care about this)
    spectrum_unit = SpectrumUnit.AMPLITUDE
    # Create a mock tag.
    tag = f"mock-{spectrum_unit.value}-spectrogram"
    return Spectrogram(
        dynamic_spectra, times, frequencies, tag, spectrum_unit, start_datetime
    )


@pytest.fixture
def base_panel(spectrogram: Spectrogram) -> BasePanel:
    """Arbitrarily create an instance of any `BasePanel` subclass, to test functionality of `BasePanel`."""
    return SpectrogramPanel(spectrogram)


@pytest.fixture
def base_time_series_panel(spectrogram: Spectrogram) -> BaseTimeSeriesPanel:
    """Arbitrarily create an instance of any `BaseTimeSeriesPanel` subclass, to test functionality of `BaseTimeSeriesPanel`."""
    return SpectrogramPanel(spectrogram)


@pytest.fixture
def frequency_cuts_panel_relative_cuts(spectrogram: Spectrogram) -> FrequencyCutsPanel:
    """A frequency cuts panel, with cuts indicated by relative times, using default keyword arguments."""
    return FrequencyCutsPanel(spectrogram, 0.0, 1.0)


@pytest.fixture
def frequency_cuts_panel_datetime_cuts(spectrogram: Spectrogram) -> FrequencyCutsPanel:
    """A frequency cuts panel, with cuts indicated by datetimes, using default keyword arguments."""
    cuts = [
        datetime.strftime(_ARBITRARY_DATETIME, TimeFormat.DATETIME),
        datetime.strftime(_ARBITRARY_DATETIME + timedelta(1), TimeFormat.DATETIME),
    ]
    return FrequencyCutsPanel(spectrogram, *cuts)


@pytest.fixture
def frequency_cuts_panel_dBb(spectrogram: Spectrogram) -> FrequencyCutsPanel:
    """A frequency cuts panel, whose spectrums are in units of decibels above the background."""
    return FrequencyCutsPanel(spectrogram, 0.0, 1.0, dBb=True)


@pytest.fixture
def frequency_cuts_panel_peak_normalised(
    spectrogram: Spectrogram,
) -> FrequencyCutsPanel:
    """A frequency cuts panel, whose spectrums are in units normalised to the peak values, respectively."""
    return FrequencyCutsPanel(spectrogram, 0.0, 1.0, peak_normalise=True)


@pytest.fixture
def time_cuts_panel(spectrogram: Spectrogram) -> TimeCutsPanel:
    """A time cuts panel, using the default keyword arguments."""
    return TimeCutsPanel(spectrogram, 90e6, 90.3e6)


@pytest.fixture
def time_cuts_panel_dBb(spectrogram: Spectrogram) -> TimeCutsPanel:
    """A time cuts panel, whose spectrums are in units of decibels above the background."""
    return TimeCutsPanel(spectrogram, 90e6, 90.3e6, dBb=True)


@pytest.fixture
def time_cuts_panel_peak_normalised(spectrogram: Spectrogram) -> TimeCutsPanel:
    """A time cuts panel, whose spectrums are in units normalised to the peak values, respectively.."""
    return TimeCutsPanel(spectrogram, 90e6, 90.3e6, peak_normalise=True)


@pytest.fixture
def integral_over_frequency_panel(
    spectrogram: Spectrogram,
) -> IntegralOverFrequencyPanel:
    """A panel plotting the integral of the spectrogram over frequency, using default system arguments."""
    return IntegralOverFrequencyPanel(spectrogram)


@pytest.fixture
def spectrogram_panel(spectrogram: Spectrogram) -> SpectrogramPanel:
    """A spectrogram panel using default keyword arguments."""
    return SpectrogramPanel(spectrogram)


@pytest.fixture
def spectrogram_panel_dBb(spectrogram: Spectrogram) -> SpectrogramPanel:
    """A spectrogram panel with dBb visualization."""
    return SpectrogramPanel(spectrogram, dBb=True)


@pytest.fixture
def spectrogram_panel_log_norm(spectrogram: Spectrogram) -> SpectrogramPanel:
    """A spectrogram panel with logarithmic normalization."""
    return SpectrogramPanel(spectrogram, log_norm=True)


@pytest.fixture
def panel_stack() -> PanelStack:
    """Create a non-interactive panel stack for testing."""
    return PanelStack(non_interactive=True)


class TestBasePanel:
    """Test shared functionality between all `BasePanel` subclasses."""

    def test_initialisation(
        self, base_panel: BasePanel, spectrogram: Spectrogram
    ) -> None:
        """Check that a `BasePanel` subclass is initialised correctly."""
        # Test the public instance attributes are set correctly
        assert base_panel.name == PanelName.SPECTROGRAM
        assert base_panel.spectrogram == spectrogram
        assert base_panel.tag == spectrogram.tag
        assert base_panel.get_identifier() is None

        with pytest.raises(ValueError):
            _ = base_panel.get_panel_format()

    def test_xaxis_hidden(self, base_panel: BasePanel, axes: Axes) -> None:
        """Check that calling for the xaxis labels to be hidden, actually hides them."""
        base_panel.set_ax(axes)
        base_panel.hide_xaxis_labels()
        assert not base_panel.get_xaxis_labels()

    def test_yaxis_hidden(self, base_panel: BasePanel, axes: Axes) -> None:
        """Check that calling for the yaxis labels to be hidden, actually hides them."""
        base_panel.set_ax(axes)
        base_panel.hide_yaxis_labels()
        assert not base_panel.get_yaxis_labels()


class TestBaseTimeSeriesPanel:
    def test_xaxis_is_time(self, base_time_series_panel: BaseTimeSeriesPanel) -> None:
        """Check that the xaxis type for a time series panel is always time."""
        assert base_time_series_panel.xaxis_type == XAxisType.TIME

    def test_times(
        self, base_time_series_panel: BaseTimeSeriesPanel, spectrogram: Spectrogram
    ) -> None:
        """Check that setting the time type for the panel, updates the times of each spectrum appropriately.

        The `spectrogram` fixture is used to construct the panel, so we use that to compare equality.
        """

        # Check both cases of time type.
        base_time_series_panel.set_time_type(TimeType.RELATIVE)
        assert np.array_equal(base_time_series_panel.times, spectrogram.times)

        base_time_series_panel.set_time_type(TimeType.DATETIMES)
        assert np.array_equal(base_time_series_panel.times, spectrogram.datetimes)

    @pytest.mark.parametrize(
        "time_type,expected_label",
        [
            (TimeType.RELATIVE, "Time [s]"),
            (TimeType.DATETIMES, "Time [UTC] (Start Date: 2025-02-13)"),
        ],
    )
    def test_xaxis_labelling(
        self,
        base_time_series_panel: BaseTimeSeriesPanel,
        axes: Axes,
        time_type: TimeType,
        expected_label: str,
    ) -> None:
        """Check axis labeling for different time types."""
        base_time_series_panel.set_ax(axes)
        base_time_series_panel.set_time_type(time_type)
        base_time_series_panel.annotate_xaxis()
        assert base_time_series_panel.get_xlabel() == expected_label


class TestFrequencyCutsPanel:
    def test_xaxis_is_frequency(
        self, frequency_cuts_panel_relative_cuts: FrequencyCutsPanel
    ) -> None:
        """Check that the xaxis type for a time series panel is always frequency.

        Arbitrarily choose the relative cuts panel, as it doesn't matter for this test.
        """
        assert frequency_cuts_panel_relative_cuts.xaxis_type == XAxisType.FREQUENCY

    def test_frequencies(
        self,
        frequency_cuts_panel_relative_cuts: FrequencyCutsPanel,
        spectrogram: Spectrogram,
    ) -> None:
        """Check that the frequencies assigned to each spectral component as managed by the panel, are equal to the frequencies
        assigned to each spectral component in the spectrogram used to construct the panel.

        Arbitrarily choose the relative cuts panel, as it doesn't matter for this test.
        """
        assert np.array_equal(
            frequency_cuts_panel_relative_cuts.frequencies, spectrogram.frequencies
        )

    def test_xaxis_annotation(
        self, frequency_cuts_panel_relative_cuts: FrequencyCutsPanel, axes: Axes
    ) -> None:
        """Check that the xaxis gets annotated appropriately."""
        frequency_cuts_panel_relative_cuts.set_ax(axes)
        frequency_cuts_panel_relative_cuts.annotate_xaxis()
        assert frequency_cuts_panel_relative_cuts.get_xlabel() == "Frequency [Hz]"

    def test_no_times_specified(self, spectrogram: Spectrogram) -> None:
        """Check that an error is raised when the class is instantiated with no times specified."""
        times: list[float] = []
        with pytest.raises(ValueError):
            FrequencyCutsPanel(spectrogram, *times)

    def test_yaxis_annotation_dBb(
        self, frequency_cuts_panel_dBb: FrequencyCutsPanel, axes: Axes
    ) -> None:
        """Check that the ylabel is annotated appropriately, in the case where the spectrum units are `dBb`."""
        frequency_cuts_panel_dBb.set_ax(axes)
        frequency_cuts_panel_dBb.annotate_yaxis()
        assert frequency_cuts_panel_dBb.get_ylabel() == "dBb"

    def test_yaxis_annotation_peak_normalised(
        self, frequency_cuts_panel_peak_normalised: FrequencyCutsPanel, axes: Axes
    ) -> None:
        """Check that the ylabel is annotated appropriately, in the case where the spectrum units are normalised to their peak values, respectively."""
        frequency_cuts_panel_peak_normalised.set_ax(axes)
        frequency_cuts_panel_peak_normalised.annotate_yaxis()
        # We don't expect any label in this case
        assert not frequency_cuts_panel_peak_normalised.get_ylabel()

    def test_yaxis_annotation(
        self,
        frequency_cuts_panel_relative_cuts: FrequencyCutsPanel,
        axes: Axes,
        spectrogram: Spectrogram,
    ) -> None:
        """Check that the ylabel is annotated appropriately, when the spectrum units are those defined by the spectrogram used to construct
        the panel.
        """
        frequency_cuts_panel_relative_cuts.set_ax(axes)
        frequency_cuts_panel_relative_cuts.annotate_yaxis()
        assert (
            frequency_cuts_panel_relative_cuts.get_ylabel()
            == f"{spectrogram.spectrum_unit.value}".capitalize()
        )


class TestTimeCutsPanel:
    def test_no_frequencies_specified(self, spectrogram: Spectrogram) -> None:
        """Check that an error is raised when the class is instantiated with no frequencies specified."""
        frequencies: list[float] = []
        with pytest.raises(ValueError):
            TimeCutsPanel(spectrogram, *frequencies)

    def test_yaxis_annotation_dBb(
        self, time_cuts_panel_dBb: TimeCutsPanel, axes: Axes
    ) -> None:
        """Check that the ylabel is annotated appropriately, in the case where the spectrum units are `dBb`."""
        # Arbitrarily set the time type (it has to be set)
        time_cuts_panel_dBb.set_time_type(TimeType.RELATIVE)
        time_cuts_panel_dBb.set_ax(axes)
        time_cuts_panel_dBb.annotate_yaxis()
        assert time_cuts_panel_dBb.get_ylabel() == "dBb"

    def test_yaxis_annotation_peak_normalised(
        self, time_cuts_panel_peak_normalised: TimeCutsPanel, axes: Axes
    ) -> None:
        """Check that the ylabel is annotated appropriately, in the case where the spectrum units are normalised to their peak values, respectively."""
        time_cuts_panel_peak_normalised.set_ax(axes)
        time_cuts_panel_peak_normalised.annotate_yaxis()
        # We don't expected any label in this case
        assert not time_cuts_panel_peak_normalised.get_ylabel()

    def test_yaxis_annotation(
        self,
        time_cuts_panel: TimeCutsPanel,
        axes: Axes,
        spectrogram: Spectrogram,
    ) -> None:
        """Check that the ylabel is annotated appropriately, when the spectrum units are those defined by the spectrogram used to construct
        the panel.
        """
        time_cuts_panel.set_ax(axes)
        time_cuts_panel.annotate_yaxis()
        assert (
            time_cuts_panel.get_ylabel()
            == f"{spectrogram.spectrum_unit.value}".capitalize()
        )


class TestIntegralOverFrequencyPanel:
    def test_yaxis_annotation(
        self, integral_over_frequency_panel: IntegralOverFrequencyPanel, axes: Axes
    ) -> None:
        """Check that the yaxis is not annotated."""
        integral_over_frequency_panel.set_ax(axes)
        integral_over_frequency_panel.annotate_yaxis()
        # We don't expect any label in this case
        assert not integral_over_frequency_panel.get_ylabel()


class TestSpectrogramPanel:
    """Test functionality specific to SpectrogramPanel."""

    def test_yaxis_annotation(
        self, spectrogram_panel: SpectrogramPanel, axes: Axes
    ) -> None:
        """Check that the yaxis is annotated appropriately."""
        spectrogram_panel.set_ax(axes)
        spectrogram_panel.annotate_yaxis()
        assert spectrogram_panel.get_ylabel() == "Frequency [Hz]"


class TestPanelStack:

    def test_time_type_getter_setters(self, panel_stack: PanelStack) -> None:
        """Check that the time type getters and setters work as expected."""
        panel_stack.time_type = TimeType.RELATIVE
        assert panel_stack.time_type == TimeType.RELATIVE

        panel_stack.time_type = TimeType.DATETIMES
        assert panel_stack.time_type == TimeType.DATETIMES

    def test_adding_panel_increments_panel_count(
        self, panel_stack: PanelStack, spectrogram_panel: SpectrogramPanel
    ) -> None:
        """Check that adding a panel increments the number of panels in the stack."""
        # Take a record of the initial number of panels
        initial_num_panels = panel_stack.num_panels

        # Check that it is zero.
        assert initial_num_panels == 0

        # Add a panel, and check that the number of panels is as expected.
        panel_stack.add_panel(spectrogram_panel)
        assert panel_stack.num_panels == initial_num_panels + 1

        # Add a panel, and check that the number of panels is as expected.
        panel_stack.add_panel(spectrogram_panel)
        assert panel_stack.num_panels == initial_num_panels + 2

    def test_adding_superimposed_panel_increments_panel_count(
        self, panel_stack: PanelStack, spectrogram_panel: SpectrogramPanel
    ) -> None:
        """Check that adding a superimposed panel increments the number of panels in the stack."""
        # Take a record of the initial number of superimposed panels
        initial_num_superimposed_panels = panel_stack.num_superimposed_panels

        # Check that it is zero.
        assert initial_num_superimposed_panels == 0

        # Add a panel, and check that the number of panels is as expected.
        panel_stack.superimpose_panel(spectrogram_panel)
        assert (
            panel_stack.num_superimposed_panels == initial_num_superimposed_panels + 1
        )

        # Add a panel, and check that the number of panels is as expected.
        panel_stack.superimpose_panel(spectrogram_panel)
        assert (
            panel_stack.num_superimposed_panels == initial_num_superimposed_panels + 2
        )

    def test_check_panel_ordering(
        self,
        panel_stack: PanelStack,
        spectrogram_panel: SpectrogramPanel,
        time_cuts_panel: TimeCutsPanel,
        frequency_cuts_panel_relative_cuts: FrequencyCutsPanel,
    ) -> None:
        """Check that panels are ordered by their `XAxisType`.

        Frequency cut panels should always be first, at the top of the plot.
        Regardless of the order in which they were added to the stack. Time series
        panels, on the other hand, should respect the order they were added to the stack.
        """
        panel_stack.add_panel(spectrogram_panel)
        panel_stack.add_panel(frequency_cuts_panel_relative_cuts)
        panel_stack.add_panel(time_cuts_panel)
        # Frequency cut panels should be first, at the top of the plot.
        assert panel_stack.panels == [
            frequency_cuts_panel_relative_cuts,
            spectrogram_panel,
            time_cuts_panel,
        ]

    def test_incompatible_time_types(
        self, panel_stack: PanelStack, spectrogram_panel: SpectrogramPanel
    ) -> None:
        """Check that the panel stack overrides the time type of the panel, if they are incompatible."""
        panel_stack.time_type = TimeType.RELATIVE
        spectrogram_panel.set_time_type(TimeType.DATETIMES)
        panel_stack.add_panel(spectrogram_panel)
        assert spectrogram_panel.get_time_type() == TimeType.RELATIVE

    def test_save_creates_file(
        self, panel_stack: PanelStack, spectrogram_panel: SpectrogramPanel
    ) -> None:
        """Test that saving a panel stack, creates a file in the filesystem with the appropriate path."""
        # Add a panel to the stack, so there's something in the figure.
        panel_stack.add_panel(spectrogram_panel)

        # Create a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Assign this to be the directory where `batches` are stored.
            pytest.MonkeyPatch().setenv("SPECTRE_DATA_DIR_PATH", temp_dir)

            # Save the figure to this temporary directory.
            file_path = panel_stack.save()

            # Check that the file path is as expected.
            assert file_path == os.path.join(
                temp_dir,
                "batches",
                "2025",
                "02",
                "13",
                "2025-02-13T06:00:00_mock-amplitude-spectrogram.png",
            )

            # Check that the file was actually created.
            assert os.path.exists(file_path)

    def test_save_no_panels(self, panel_stack: PanelStack) -> None:
        """Check that trying to save a plot with no panels raises an error."""
        with pytest.raises(ValueError):
            panel_stack.save()

    def test_show_no_panels(self, panel_stack: PanelStack) -> None:
        """Check that trying to show a plot with no panels raises an error."""
        with pytest.raises(ValueError):
            panel_stack.show()



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Callable, cast, Optional
from logging import getLogger

_LOGGER = getLogger(__name__)

from spectre_core.capture_configs import (
    CaptureTemplate,
    CaptureMode,
    Parameters,
    Bound,
    PName,
    OneOf,
    get_base_capture_template,
    get_base_ptemplate,
    validate_fixed_center_frequency,
    validate_swept_center_frequency,
)

from .._receiver import Receiver, ReceiverName
from .._specs import SpecName

LOW_IF_SAMPLE_RATE_CUTOFF = 2e6
LOW_IF_PERMITTED_SAMPLE_RATES = [LOW_IF_SAMPLE_RATE_CUTOFF / (2**i) for i in range(6)]


class SDRplayReceiver(ABC, Receiver):
    """An abstract base class for SDRplay devices."""

    def __init__(self, name: ReceiverName, mode: Optional[str]):
        """Initialise an instance of an `SDRplayReceiver`."""
        super().__init__(name, mode)

        self.add_spec(SpecName.SAMPLE_RATE_LOWER_BOUND, 62.5e3)
        self.add_spec(SpecName.SAMPLE_RATE_UPPER_BOUND, 10.66e6)
        self.add_spec(SpecName.FREQUENCY_LOWER_BOUND, 1e3)
        self.add_spec(SpecName.FREQUENCY_UPPER_BOUND, 2e9)
        self.add_spec(SpecName.IF_GAIN_UPPER_BOUND, -20)
        self.add_spec(SpecName.IF_GAIN_LOWER_BOUND, -59)
        self.add_spec(SpecName.API_RETUNING_LATENCY, 25 * 1e-3)

        # bandwidth == 0 means 'AUTO', i.e. the largest bandwidth compatible with the sample rate
        self.add_spec(
            SpecName.BANDWIDTH_OPTIONS,
            [0, 200e3, 300e3, 600e3, 1.536e6, 5e6, 6e6, 7e6, 8e6],
        )

    @abstractmethod
    def get_rf_gains(self, center_frequency: float) -> list[int]:
        """Get an ordered list of RF gain values corresponding to each LNA state at the specified center frequency.

        The values are taken from the gain reduction tables documented in the SDRplay API specification, and are
        unique to each model. Note that negative gain values represent positive gain reduction.
        """


def _validate_rf_gain(rf_gain: int, expected_rf_gains: list[int]):
    """Validate the RF gain value against the expected values for the current LNA state.

    The RF gain is determined by the LNA state and can only take specific values as documented in the
    gain reduction tables of the SDRplay API specification.

    For implementation details, refer to the `gr-sdrplay3` OOT module:
    https://github.com/fventuri/gr-sdrplay3/blob/v3.11.0.9/lib/rsp_impl.cc#L378-L387
    """
    if rf_gain not in expected_rf_gains:
        raise ValueError(
            f"The value of RF gain must be one of {expected_rf_gains}. "
            f"Got {rf_gain}."
        )


def _validate_low_if_sample_rate(sample_rate: int) -> None:
    """Validate the sample rate if the receiver is operating in low IF mode.

    The minimum physical sampling rate of the SDRplay hardware is 2 MHz. Lower effective rates can be achieved
    through decimation, as handled by the `gr-sdrplay3` OOT module. This function ensures that the sample rate
    is not silently adjusted by the backend.

    For implementation details, refer to:
    https://github.com/fventuri/gr-sdrplay3/blob/v3.11.0.9/lib/rsp_impl.cc#L140-L179
    """
    if sample_rate <= LOW_IF_SAMPLE_RATE_CUTOFF:
        if sample_rate not in LOW_IF_PERMITTED_SAMPLE_RATES:
            raise ValueError(
                f"If the requested sample rate is less than or equal to {LOW_IF_SAMPLE_RATE_CUTOFF}, "
                f"the receiver will be operating in low IF mode. "
                f"So, the sample rate must be exactly one of {LOW_IF_PERMITTED_SAMPLE_RATES}. "
                f"Got sample rate {sample_rate} Hz"
            )


def make_pvalidator_fixed_center_frequency(
    receiver: SDRplayReceiver,
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_fixed_center_frequency(parameters)

        # Validate the sample rate, in the case the receiver will be operating in low if mode.
        sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
        _validate_low_if_sample_rate(sample_rate)

        # Validate the rf gain value, which is a function of center frequency.
        rf_gain = cast(int, parameters.get_parameter_value(PName.RF_GAIN))
        center_frequency = cast(
            float, parameters.get_parameter_value(PName.CENTER_FREQUENCY)
        )
        _validate_rf_gain(rf_gain, receiver.get_rf_gains(center_frequency))

    return pvalidator


def make_pvalidator_swept_center_frequency(
    receiver: SDRplayReceiver,
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_swept_center_frequency(
            parameters, receiver.get_spec(SpecName.API_RETUNING_LATENCY)
        )

        # Validate the sample rate, in the case the receiver will be operating in low if mode.
        sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
        _validate_low_if_sample_rate(sample_rate)

        # Validate the rf gain value, which is a function of center frequency.
        rf_gain = cast(int, parameters.get_parameter_value(PName.RF_GAIN))
        min_frequency = cast(float, parameters.get_parameter_value(PName.MIN_FREQUENCY))
        max_frequency = cast(float, parameters.get_parameter_value(PName.MAX_FREQUENCY))
        _validate_rf_gain(rf_gain, receiver.get_rf_gains(min_frequency))
        _validate_rf_gain(rf_gain, receiver.get_rf_gains(max_frequency))

        # Check that we are not cross a threshold where the LNA state has to change
        if receiver.get_rf_gains(min_frequency) != receiver.get_rf_gains(max_frequency):
            _LOGGER.warning(
                "Crossing a threshold where the LNA state has to change. Performance may be reduced."
            )

    return pvalidator


def make_capture_template_fixed_center_frequency(
    receiver: SDRplayReceiver,
) -> CaptureTemplate:

    capture_template = get_base_capture_template(CaptureMode.FIXED_CENTER_FREQUENCY)
    capture_template.add_ptemplate(get_base_ptemplate(PName.BANDWIDTH))
    capture_template.add_ptemplate(get_base_ptemplate(PName.IF_GAIN))
    capture_template.add_ptemplate(get_base_ptemplate(PName.RF_GAIN))

    capture_template.set_defaults(
        (PName.BATCH_SIZE, 3.0),
        (PName.CENTER_FREQUENCY, 95800000),
        (PName.SAMPLE_RATE, 500000),
        (PName.BANDWIDTH, 300000),
        (PName.WINDOW_HOP, 512),
        (PName.WINDOW_SIZE, 1024),
        (PName.WINDOW_TYPE, "blackman"),
        (PName.RF_GAIN, 0),
        (PName.IF_GAIN, -30),
    )

    capture_template.add_pconstraint(
        PName.CENTER_FREQUENCY,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH, [OneOf(receiver.get_spec(SpecName.BANDWIDTH_OPTIONS))]
    )
    capture_template.add_pconstraint(
        PName.IF_GAIN,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.IF_GAIN_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.IF_GAIN_UPPER_BOUND),
            )
        ],
    )
    return capture_template


def make_capture_template_swept_center_frequency(
    receiver: Receiver,
) -> CaptureTemplate:

    capture_template = get_base_capture_template(CaptureMode.SWEPT_CENTER_FREQUENCY)
    capture_template.add_ptemplate(get_base_ptemplate(PName.BANDWIDTH))
    capture_template.add_ptemplate(get_base_ptemplate(PName.IF_GAIN))
    capture_template.add_ptemplate(get_base_ptemplate(PName.RF_GAIN))

    capture_template.set_defaults(
        (PName.BATCH_SIZE, 3.0),
        (PName.MIN_FREQUENCY, 95000000),
        (PName.MAX_FREQUENCY, 100000000),
        (PName.SAMPLES_PER_STEP, 80000),
        (PName.FREQUENCY_STEP, 2000000),
        (PName.SAMPLE_RATE, 2e6),
        (PName.BANDWIDTH, 1536000),
        (PName.WINDOW_HOP, 512),
        (PName.WINDOW_SIZE, 1024),
        (PName.WINDOW_TYPE, "blackman"),
        (PName.RF_GAIN, 0),
        (PName.IF_GAIN, -30),
    )

    capture_template.add_pconstraint(
        PName.MIN_FREQUENCY,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.MAX_FREQUENCY,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND),
            )
        ],
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH, [OneOf(receiver.get_spec(SpecName.BANDWIDTH_OPTIONS))]
    )
    capture_template.add_pconstraint(
        PName.IF_GAIN,
        [
            Bound(
                lower_bound=receiver.get_spec(SpecName.IF_GAIN_LOWER_BOUND),
                upper_bound=receiver.get_spec(SpecName.IF_GAIN_UPPER_BOUND),
            )
        ],
    )
    return capture_template



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import time

from gnuradio import spectre
from gnuradio import uhd

from spectre_core.capture_configs import Parameters, PName
from spectre_core.config import get_batches_dir_path
from ._gr import spectre_top_block


class fixed_center_frequency(spectre_top_block):
    def flowgraph(self, tag: str, parameters: Parameters) -> None:

        # Unpack capture config parameters
        sample_rate = parameters.get_parameter_value(PName.SAMPLE_RATE)
        gain = parameters.get_parameter_value(PName.GAIN)
        center_freq = parameters.get_parameter_value(PName.CENTER_FREQUENCY)
        master_clock_rate = parameters.get_parameter_value(PName.MASTER_CLOCK_RATE)
        wire_format = parameters.get_parameter_value(PName.WIRE_FORMAT)
        batch_size = parameters.get_parameter_value(PName.BATCH_SIZE)
        bandwidth = parameters.get_parameter_value(PName.BANDWIDTH)

        # Blocks
        master_clock_rate = f"master_clock_rate={master_clock_rate}"
        self.uhd_usrp_source = uhd.usrp_source(
            ",".join(("", "", master_clock_rate)),
            uhd.stream_args(
                cpu_format="fc32",
                otw_format=wire_format,
                args="",
                channels=[0],
            ),
        )
        self.uhd_usrp_source.set_samp_rate(sample_rate)
        self.uhd_usrp_source.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)

        self.uhd_usrp_source.set_center_freq(center_freq, 0)
        self.uhd_usrp_source.set_antenna("RX2", 0)
        self.uhd_usrp_source.set_bandwidth(bandwidth, 0)
        self.uhd_usrp_source.set_rx_agc(False, 0)
        self.uhd_usrp_source.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source.set_gain(gain, 0)
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            get_batches_dir_path(), tag, batch_size, sample_rate, False, "rx_freq", 0
        )

        # Connections
        self.connect((self.uhd_usrp_source, 0), (self.spectre_batched_file_sink, 0))


class swept_center_frequency(spectre_top_block):
    def flowgraph(self, tag: str, parameters: Parameters) -> None:
        # Unpack capture config parameters
        sample_rate = parameters.get_parameter_value(PName.SAMPLE_RATE)
        bandwidth = parameters.get_parameter_value(PName.BANDWIDTH)
        min_frequency = parameters.get_parameter_value(PName.MIN_FREQUENCY)
        max_frequency = parameters.get_parameter_value(PName.MAX_FREQUENCY)
        frequency_step = parameters.get_parameter_value(PName.FREQUENCY_STEP)
        samples_per_step = parameters.get_parameter_value(PName.SAMPLES_PER_STEP)
        master_clock_rate = parameters.get_parameter_value(PName.MASTER_CLOCK_RATE)
        master_clock_rate = master_clock_rate = parameters.get_parameter_value(
            PName.MASTER_CLOCK_RATE
        )
        wire_format = parameters.get_parameter_value(PName.WIRE_FORMAT)
        gain = parameters.get_parameter_value(PName.GAIN)
        batch_size = parameters.get_parameter_value(PName.BATCH_SIZE)

        master_clock_rate = f"master_clock_rate={master_clock_rate}"
        self.uhd_usrp_source = uhd.usrp_source(
            ",".join(("", "", master_clock_rate)),
            uhd.stream_args(
                cpu_format="fc32",
                otw_format=wire_format,
                args="",
                channels=[0],
            ),
        )
        self.uhd_usrp_source.set_samp_rate(sample_rate)
        self.uhd_usrp_source.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)
        self.uhd_usrp_source.set_center_freq(min_frequency, 0)
        self.uhd_usrp_source.set_antenna("RX2", 0)
        self.uhd_usrp_source.set_bandwidth(bandwidth, 0)
        self.uhd_usrp_source.set_rx_agc(False, 0)
        self.uhd_usrp_source.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source.set_gain(gain, 0)

        self.spectre_sweep_driver = spectre.sweep_driver(
            min_frequency,
            max_frequency,
            frequency_step,
            sample_rate,
            samples_per_step,
            "freq",
        )

        self.spectre_batched_file_sink = spectre.batched_file_sink(
            get_batches_dir_path(),
            tag,
            batch_size,
            sample_rate,
            True,
            "rx_freq",
            min_frequency,
        )

        # Connections
        self.msg_connect(
            (self.spectre_sweep_driver, "retune_command"),
            (self.uhd_usrp_source, "command"),
        )
        self.connect((self.uhd_usrp_source, 0), (self.spectre_batched_file_sink, 0))
        self.connect((self.uhd_usrp_source, 0), (self.spectre_sweep_driver, 0))



# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Real-time, extensible post-processing of SDR data into spectrograms."""

from .plugins._fixed_center_frequency import FixedEventHandler
from .plugins._swept_center_frequency import SweptEventHandler

from ._factory import get_event_handler, get_event_handler_cls_from_tag
from ._post_processor import start_post_processor

__all__ = [
    "FixedEventHandler",
    "SweptEventHandler",
    "start_post_processor",
    "get_event_handler",
    "get_event_handler_cls_from_tag",
]
